import os
import glob
import yaml
import json
from itertools import repeat
from typing import Dict, Optional, List, Union

import click
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_pdf import PdfPages
from quickstats.utils.process_tools import parallel_run
from quickstats.plots.template import draw_ATLAS_label, format_axis_ticks, parse_transform, draw_hatches, \
                                      draw_sigma_bands, draw_sigma_lines, get_box_dimension, draw_text, centralize_axis
from quickstats.components.numerics import ceildiv, pretty_float
from quickstats.utils.common_utils import json_load

BASE_STYLE = {
    'pull': {
        'marker'     : 'o', 
        'markersize' :  6,
        'ls'         : 'none',
        'color'      : 'k',
        'capsize'    : 4,
        'capthick'   : 2, 
        'linewidth'  : 2
    },
    'one_sigma':{
        'marker'     : None, 
        'markersize' : 0,
        'ls'         : 'none',
        'color'      : 'r',
        'capsize'    : 4,
        'capthick'   : 2, 
        'linewidth'  : 2.5
    },    
    'shade':{
        'hatch' : 'xxxxx', 
        'color' : 'gray', 
        'alpha' : 0.1,
        'fill'  : True
    },
    'legend': {
        'x': 0.48,
        'dx': 0.01,
        'dy': 0.01,
        'fontsize': 15,
        'fontsize_minor': 12,
        'box_width': 0.06,
        'box_height': 0.025,
        'line_width': 0.05
    },
    'sigma_lines':{
        'color'    : 'gray',
        'linestyle': '--'
    },
    'logo':{
        'x': 0.02,
        'fontsize': 25
    },    
    'label':{
        'x': 0.02,
        'dy': 0.01,
        'fontsize': 15
    },
    'figure': {
        'width' : 8,
        'height': 14,
        'dpi'   : 100
    },
    'xaxis' :{
        'labelsize': 15
    },
    'yaxis':{
        'labelsize': 12
    },
    'xaxis_label':{
        'fontsize': 18,
        'loc' : 'right'
    }
}

DEFAULT_STYLE = {
    'prefit_1': {
        'linewidth': 2,
        'edgecolor': '#000155',
        'alpha': 1,
        'hatch': None,
        'fill': False,
    },
    'prefit_2':{
        'linewidth': 2,
        'edgecolor': '#027734',
        'alpha': 1,
        'hatch': None,
        'fill': False,
    },
    'postfit_1':{
        'linewidth': 0,
        'color': '#0094ff',
        'alpha': 1,
        'hatch': None,
        'fill': True,
    },
    'postfit_2':{
        'linewidth': 0,
        'color': "#44ff94",
        'alpha': 1,
        'hatch': None,
        'fill': True,
    }, **BASE_STYLE
}
        
TREX_STYLE = {
    'prefit_1': {
        'linewidth': 2,
        'color': '#00FFFF',
        'edgecolor': '#00FFFF',
        'alpha': 0.6,
        'hatch': None,
        'fill': False,
    },
    'prefit_2':{
        'linewidth': 2,
        'color': '#6495ED',
        'edgecolor': '#6495ED',
        'alpha': 0.6,
        'hatch': None,
        'fill': False,
    },
    'postfit_1':{
        'linewidth': 0,
        'color': '#00FFFF',
        'alpha': 0.6,
        'hatch': None,
        'fill': True,
    },
    'postfit_2':{
        'linewidth': 0,
        'color': "#6495ED",
        'alpha': 0.6,
        'hatch': None,
        'fill': True,
    }, **BASE_STYLE
}

class NPRankingPlot:
    def __init__(self, input_dir:str=None, poi:Optional[str]=None):
        self._data = None
        if input_dir is not None:
            self.load(input_dir, poi=poi)
        
    @property
    def data(self):
        return self._data
    
    @staticmethod
    def _extract_pull_result(file:str, poi:str=None):
        with open(file, 'r') as infile:
                data = json_load(infile)
        result = {}
        nuis_data = data['nuis']
        result['nuis_label']  = nuis_data['nuisance']
        result['nuis_val']    = nuis_data['nuis_hat'] - nuis_data['nuis_nom']
        result['nuis_up']     = nuis_data['nuis_hi']
        result['nuis_down']   = nuis_data['nuis_lo']
        result['nuis_prefit'] = nuis_data['nuis_prefit']
        if poi:
            poi_data = data['pois'].get(poi, None)
            if not poi_data:
                raise ValueError("No impact evaluated for the POI \"{}\" in the file {}".format(poi, file))
            result['poi_hat']      = poi_data['hat']
            result['impact_postfit_up']    = poi_data['up'] 
            result['impact_postfit_down']  = poi_data['down'] 
            result['impact_prefit_up']   = poi_data['up_nom'] 
            result['impact_prefit_down'] = poi_data['down_nom']
        return result
    
    @staticmethod
    def extract_pull_results(input_dir:str, poi:str=None):
        if not os.path.exists(input_dir):
            raise FileNotFoundError('input directory {} does not exist'.format(input_dir))
        if not os.path.isdir(input_dir):
            raise ValueError('{} is not a directory'.format(input_dir))
        input_files = glob.glob(os.path.join(input_dir, "*.json"))
        data = parallel_run(NPRankingPlot._extract_pull_result, input_files, repeat(poi))
        df = pd.DataFrame(data)
        if poi:
            df['impact_postfit_up']   -= df['poi_hat']
            df['impact_postfit_down'] -= df['poi_hat']
            df['impact_prefit_up']    -= df['poi_hat']
            df['impact_prefit_down']  -= df['poi_hat']
            df['impact_postfit_total'] = abs(df['impact_postfit_up']) + abs(df['impact_postfit_down'])
            df['impact_prefit_total']  = abs(df['impact_prefit_up']) + abs(df['impact_prefit_down'])
        return df
    
    def load(self, input_dir:str, poi:Optional[str]=None):
        self._data = self.extract_pull_results(input_dir, poi=poi)
        
    @staticmethod
    def get_processed_data(df, num=None, threshold=None, ranking=True):
        processed_df = df.copy()
        if 'impact_postfit_total' in processed_df:
            if ranking:
                processed_df.sort_values('impact_postfit_total', ascending=False, inplace=True)
            if threshold is not None:
                processed_df = processed_df[processed_df['impact_postfit_total'] >= threshold]            
        if num is not None:
            processed_df = processed_df.head(num)
        processed_df = processed_df.iloc[::-1]
        return processed_df
    
    @staticmethod
    def setup_axis(ax1, ax2, **styles):
        ax2.set_zorder(ax1.get_zorder()-1)
        ax1.patch.set_visible(False)
        ax1.tick_params(axis='y', which='both', left=False, right=False, **styles['yaxis'])
        ax1.tick_params(axis='x', direction="in", **styles['xaxis'])
        ax2.tick_params(axis='x', direction="in", **styles['xaxis'])    
        
    @staticmethod
    def draw_impact_legend(axis, poi:str, impact_type:str, x:float, y:float, 
                           label_1:str, label_2:str, dx:float=0.01, dy:float=0.01,
                           box_width:float=0.06, box_height:float=0.025, **styles):
        """
            Draw legend for prefit/postfit impact.
            x: starting x-position in Axes coordinates
            y: starting y-position in Axes coordinates
        """
        header = '{} Impact on {}:'.format(impact_type.title(), poi)
        fontsize, fontsize_minor = styles['legend']['fontsize'], styles['legend']['fontsize_minor']
        style_type1, style_type2 = styles['{}_1'.format(impact_type)], styles['{}_2'.format(impact_type)]
        text_prefit = axis.text(x, y - dy, header, fontsize=fontsize, transform=axis.transAxes,
                                horizontalalignment='left', verticalalignment='top')
        _, _, ymin, _ = get_box_dimension(text_prefit)
        # y-position of marker and label
        y_marker = ymin - dy - box_height
        y_label  = y_marker + box_height/2
        # draw first marker
        axis.add_patch(Rectangle((x + dx, y_marker), box_width, 
                                 box_height, transform=axis.transAxes, **style_type1))
        # draw label for first marker
        label_box_1 = axis.text(x + 2*dx + box_width, y_label, label_1, 
                                  fontsize=fontsize_minor, transform=axis.transAxes,
                                  horizontalalignment='left',
                                  verticalalignment='center')
        _, x_2, _, _ = get_box_dimension(label_box_1)
        # draw second marker
        axis.add_patch(Rectangle((x_2 + 2*dx, y_marker), box_width, 
                                 box_height, transform=axis.transAxes, **style_type2))
        # draw label for second marker
        label_box_2 = axis.text(x_2 + 3*dx + box_width, y_label, label_2, 
                                  fontsize=fontsize_minor, transform=axis.transAxes,
                                  horizontalalignment='left',
                                  verticalalignment='center')
        return y_marker
    
    @staticmethod
    def draw_legend(axis, poi:str=r"$\mu$", x:float=0.48, y:float=-1.2, show_sigma:bool=True, show_prefit:bool=True,
                    show_postfit:bool=True, correlation:bool=True, **styles):
        """
            Draw legends.
            x: starting x-position in Axes coordinates
            y: starting y-position in Data coordinates
        """
        dx = styles['legend']['dx']
        dy = styles['legend']['dy']
        line_width     = styles['legend']['line_width']
        box_width      = styles['legend']['box_width']
        box_height     = styles['legend']['box_height']
        fontsize       = styles['legend']['fontsize']
        fontsize_minor = styles['legend']['fontsize_minor']
        # starting x-position of pull and one sigma legend markers
        x_marker = x + dx + line_width
        # current y-position
        y_curr  = y
        ########################################################################################
        # draw pull legend
        pull_styles = {k:v for k,v in styles['pull'].items() if 'cap' not in k}
        transform = transforms.blended_transform_factory(axis.transAxes, axis.transData) 
        axis.errorbar(x_marker, y_curr, yerr=None, xerr=line_width, 
                      transform=transform, **pull_styles)
        text_pull = axis.text(x_marker + line_width + 2*dx, y, 'Nuis. Param. Pull', 
                              fontsize=fontsize, transform=transform,
                              horizontalalignment='left',
                              verticalalignment='center')
        #########################################################################################
        
        #########################################################################################
        # draw one sigma legend
        _, _, ymin, ymax = get_box_dimension(text_pull)
        y_curr = ymin
        if show_sigma:
            height = ymax - ymin
            one_sigma_styles = {k:v for k,v in styles['one_sigma'].items() if 'cap' not in k}
            axis.errorbar(x_marker, ymin - height, yerr=None, xerr=line_width, 
                          transform=axis.transAxes, **one_sigma_styles)
            text_sigma = axis.text(x_marker + line_width + 2*dx, ymin - height, '1 standard deviation', 
                                   fontsize=fontsize, transform=axis.transAxes,
                                   horizontalalignment='left',
                                   verticalalignment='center')
            _, _, y_curr, _ = get_box_dimension(text_sigma)
        #########################################################################################
        if correlation:
            prefit_label1 = r"Correlated"
            prefit_label2 = r"Anticorrelated"
            postfit_label1 = r"Correlated"
            postfit_label2 = r"Anticorrelated"
        else:  
            prefit_label1 = r"$\theta = \hat{\theta}+\Delta\theta$"
            prefit_label2 = r"$\theta = \hat{\theta}-\Delta\theta$"
            postfit_label1 = r"$\theta = \hat{\theta}+\Delta\hat{\theta}$"
            postfit_label2 = r"$\theta = \hat{\theta}-\Delta\hat{\theta}$"
        #########################################################################################
        # draw prefit impact legend
        if show_prefit:
            y_curr = NPRankingPlot.draw_impact_legend(axis, poi, "prefit", x, y_curr, prefit_label1, 
                                                      prefit_label2, dx, dy, box_width, box_height, **styles)
        #########################################################################################
        #########################################################################################
        # draw postfit impact legend
        if show_postfit:
            y_curr = NPRankingPlot.draw_impact_legend(axis, poi, "postfit", x, y_curr, postfit_label1, 
                                                      postfit_label2, dx, dy, box_width, box_height, **styles)
        #########################################################################################
        
    @staticmethod
    def plot_pulls(axis, data, height=1.0, show_sigma:bool=True, **styles):
        n = len(data)
        y = np.arange(0, n*height, height)
        axis.errorbar(data['nuis_val'], y, yerr=None, 
                      xerr=[abs(data['nuis_down']), abs(data['nuis_up'])], 
                      zorder=3.1, **styles['pull'])
        if show_sigma:
            axis.errorbar(data['nuis_val'], y, yerr=None, 
                          xerr=[abs(data['nuis_prefit']), abs(data['nuis_prefit'])],
                          zorder=3.0, **styles['one_sigma'])        
        
    @staticmethod
    def plot_impact(axis, data, show_prefit:bool=True, show_postfit:bool=True, correlation:bool=True, onesided:bool=True, 
                    relative:bool=False, height:float=1.0, spacing:float=0., **styles):     
        n = len(data)  
        impact_types = []
        if show_postfit:
            impact_types.append('postfit')
        if show_prefit:
            impact_types.append('prefit')
        if not correlation:
            impact_types = impact_types[::-1]
        for it in impact_types:
            prefix = 'impact_{}'.format(it)
            # retrieve the impact data
            if relative:
                impact_down = data['{}_down'.format(prefix)].values / data['poi_hat'].values
                impact_up = data['{}_up'.format(prefix)].values / data['poi_hat'].values           
            else:
                impact_down = data['{}_down'.format(prefix)].values
                impact_up = data['{}_up'.format(prefix)].values

            if correlation:
                # negative-valued correlated impact 
                neg_type1 = (impact_down <= 0)
                # positive-valued correlated imapact
                pos_type1 = (impact_up >= 0)
                # negative-valued anticorrelated impact
                neg_type2 = ~pos_type1
                # positive-valued anticorrelated impact
                pos_type2 = ~neg_type1
                x_neg_type1 = impact_down[neg_type1]
                x_pos_type1 = impact_up[pos_type1]
                x_neg_type2 = impact_up[neg_type2]
                x_pos_type2 = impact_down[pos_type2]
            else:
                # negative-valued +1 sigma impact
                neg_type1 = impact_up < 0
                # positive-valued +1 sigma impact
                pos_type1 = impact_up > 0
                # negative-valued -1 sigma impact
                neg_type2 = impact_down < 0
                # positive-valued -1 sigma impact
                pos_type2 = impact_down > 0 
                x_neg_type1 = impact_up[neg_type1]
                x_pos_type1 = impact_up[pos_type1]
                x_neg_type2 = impact_down[neg_type2]    
                x_pos_type2 = impact_down[pos_type2]

            style_type1, style_type2 = styles['{}_1'.format(it)], styles['{}_2'.format(it)]
            x = np.zeros(n)
            y_type1 = np.arange(0, n*height, height) - height/2 + spacing/2
            y_type2 = y_type1.copy()
            w = height - spacing
            h = np.full((n), w, dtype=float)
            if onesided:
                all_neg = (impact_down < 0) & (impact_up < 0)
                all_pos = (impact_down > 0) & (impact_up > 0)
                h[all_neg] = w/2
                h[all_pos] = w/2
                # secondary bar
                # put type1 bar above by default and type2 bar below
                y_type1[all_neg] = y_type1[all_neg] + w/2
                y_type1[all_pos] = y_type1[all_pos] + w/2

            # negative-valued type 1
            axis.bar(x_neg_type1, height=h[neg_type1], 
                     width=abs(x_neg_type1), bottom=y_type1[neg_type1], 
                     align='edge', **style_type1)
            # positive-valued type 1
            axis.bar(x[pos_type1], height=h[pos_type1], 
                     width=x_pos_type1, bottom=y_type1[pos_type1], 
                     align='edge', **style_type1)          
            # negative-valued type 2
            axis.bar(x_neg_type2, height=h[neg_type2], 
                     width=abs(x_neg_type2), bottom=y_type2[neg_type2], 
                     align='edge', **style_type2)
            # positive-valued type 2
            axis.bar(x[pos_type2], height=h[pos_type2], 
                     width=x_pos_type2, bottom=y_type2[pos_type2], 
                     align='edge', **style_type2)
            
    @staticmethod
    def draw_np_labels(axis, data, padding:int=7, height:float=1.0):
        n = len(data)
        y_ticks = np.arange(-padding, height*(n+1), height)
        y_ticks_label = np.concatenate((['']*padding, data['nuis_label'], ['']*1))
        axis.set_yticks(y_ticks)
        axis.set_yticklabels(y_ticks_label)
        axis.set_ylim((-padding, n + height/2))
    
    def draw_labels(axis, extra_text:str=None, elumi_label:bool=True, ranking_label:bool=True, height=1.0,
                    energy:float=13, lumi:float=139, r_start:int=0, r_end:int=0, **styles):
        """
            extra_text: extra labels below the ATLAS label, use "//" as newline delimiter
        """
        x_logo = styles['logo']['x']
        fontsize_logo = styles['logo']['fontsize']
        _, _, y_curr, _ = draw_ATLAS_label(axis, x=x_logo, y=-height, transform_y='data', fontsize=fontsize_logo)
        dy = styles['label']['dy']        
        x_label = styles['label']['x']
        text_options = {
            'fontsize'  : styles['label']['fontsize'],
            'transform' : axis.transAxes,
            'horizontalalignment': 'left',
            'verticalalignment': 'top'
        }
        texts = []
        if extra_text is not None:
            texts += extra_text.split("//")
        if elumi_label:
            energy = pretty_float(energy)
            lumi = pretty_float(lumi)
            s = r'$\sqrt{s}$' + ' = {} TeV, {} '.format(energy, lumi) + r'fb$^{-1}$'
            texts.append(s)
        if ranking_label:
            s = 'Rank {} to {}'.format(r_start, r_end)
            texts.append(s)
        for text in texts:
            _, _, y_curr, _ = draw_text(axis, x_label, y_curr - dy, text, **text_options)
    @staticmethod
    def parse_style(style:str):
        if os.path.exists(style):
            return yaml.safe_load(open(style))
        if isinstance(style, str):
            if style.lower() == 'default':
                return DEFAULT_STYLE
            elif style.lower() == 'trex':
                return TREX_STYLE
            else:
                raise ValueError('unknown style: {}'.format(style))
        return DEFAULT_STYLE
    
    @staticmethod
    def _plot(data, show_sigma:bool=True, show_prefit:bool=True, show_postfit:bool=True, sigma_bands:bool=False,
              sigma_lines:bool=True, shade:bool=True, correlation:bool=True, onesided:bool=True, relative:bool=False,
              theta_max:float=2, padding:int=7, height:float=1, spacing:float=0, display_poi:str=r"$\mu$", 
              extra_text:str=None, elumi_label:bool=True, ranking_label:bool=True, energy:float=13, lumi:float=139, 
              r_start:int=1, style:Dict=DEFAULT_STYLE):
        # setup figure and axes
        plt.clf()
        fig_width, fig_height = style['figure']['width'], style['figure']['height']
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax1 = fig.add_subplot(111)
        ax2 = ax1.twiny()
        NPRankingPlot.setup_axis(ax1, ax2, **style)
        ax1.set_xlim((-theta_max, theta_max))
        ax1.set_xlabel(r"$(\hat{\theta}-\theta_0)/\Delta\theta$" , **style['xaxis_label'])
        if relative:
            ax2.set_xlabel(r"$\Delta$ " + display_poi +"/" + display_poi, 
                           **style['xaxis_label'], labelpad=10)
        else:
            ax2.set_xlabel(r"$\Delta$ " + display_poi , **style['xaxis_label'], labelpad=10)
        if sigma_bands:
            shade, show_prefit, show_postfit = False, False, False
        n = len(data)
        NPRankingPlot.plot_impact(ax2, data, show_prefit=show_prefit, show_postfit=show_postfit,
                                  correlation=correlation, onesided=onesided, relative=relative,
                                  height=height, spacing=spacing, **style)
        centralize_axis(ax2, 'x')
        if shade:
            draw_hatches(ax2, n, height=height, **style['shade'])
        NPRankingPlot.plot_pulls(ax1, data, height=height, **style)
        if sigma_bands:
            draw_sigma_bands(ax1, n, height=height)
        if sigma_lines:
            draw_sigma_lines(ax1, n, height=height, **style['sigma_lines'])
        NPRankingPlot.draw_np_labels(ax2, data, padding=padding, height=height)
        format_axis_ticks(ax1, y_axis=False, x_axis_styles=style['xaxis'])
        format_axis_ticks(ax2, y_axis=False, x_axis_styles=style['xaxis'])
        impact_xlim = ax2.get_xlim()
        if (abs(impact_xlim[0]) + abs(impact_xlim[1])) < 0.05:
            ax2.xaxis.set_major_locator(plt.MaxNLocator(3))        
        NPRankingPlot.draw_labels(ax1, extra_text=extra_text, height=height, elumi_label=elumi_label,
                                  ranking_label=ranking_label, energy=energy, lumi=lumi, 
                                  r_start=r_start, r_end=r_start + n - 1, **style)
        x_legend = style['legend']['x']
        NPRankingPlot.draw_legend(ax1, display_poi, x=x_legend, y=-1.2*height, show_prefit=show_prefit,
                                  show_postfit=show_postfit, correlation=correlation, **style)
        return fig
    
    def plot(self, show_sigma:bool=True, show_prefit:bool=True, show_postfit:bool=True, sigma_bands:bool=False,
             sigma_lines:bool=True, shade:bool=True, correlation:bool=True, onesided:bool=True, relative:bool=False,
             theta_max:float=2, padding:int=7, height:float=1, spacing:float=0, display_poi:str=r"$\mu$", 
             extra_text:str=None, elumi_label:bool=True, ranking_label:bool=True, energy:float=13, lumi:float=139,
             n_rank=None, rank_per_plot=20, combine_pdf=True, threshold=0, ranking=True, fix_axis_scale=True,
             outdir="ranking_plots", outname:str='ranking', style:Union[str, Dict]=DEFAULT_STYLE):
        if isinstance(style, str):
            style = self.parse_style(style)
        if self.data is None:
            raise ValueError("no data to process")
        has_impact = any('impact' in i for i in self.data)
        if not has_impact:
            ranking, threshold, show_prefit, show_postfit = False, False, False, False           
        processed_data = self.get_processed_data(self.data, num=n_rank, threshold=threshold, ranking=ranking)
        num_data = len(processed_data)
        if not num_data:
            print("INFO: No data passing threshold. Skipped.")
        n_plot = ceildiv(num_data, rank_per_plot)
        dpi = style['figure']['dpi']
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        if combine_pdf:
            pdf_name = os.path.join(outdir, '{}_rank_{:04d}_to_{:04d}.pdf'.format(outname, 1, num_data))
            pp = PdfPages(pdf_name)
        nominal_xlim = None
        for i in range(n_plot):
            start, end = rank_per_plot*i, rank_per_plot*(i+1)
            data_slice = processed_data[::-1][start:end][::-1]
            fig = self._plot(data_slice, show_sigma=show_sigma, show_prefit=show_prefit, show_postfit=show_postfit, 
                             sigma_bands=sigma_bands, sigma_lines=sigma_lines, shade=shade, correlation=correlation,
                             onesided=onesided, relative=relative, theta_max=theta_max, padding=padding, height=height, 
                             spacing=spacing, display_poi=display_poi, extra_text=extra_text, elumi_label=elumi_label,
                             ranking_label=ranking_label, energy=energy, lumi=lumi, r_start=start+1, style=style)
            if fix_axis_scale:
                if nominal_xlim is None:
                    nominal_xlim = fig.get_axes()[1].get_xlim()
                else:
                    fig.get_axes()[1].set_xlim(nominal_xlim)
            base_name = os.path.join(outdir, '{}_rank_{:04d}_to_{:04d}'.format(outname, start+1, start+len(data_slice)))
            fig.savefig(base_name+'.eps', bbox_inches='tight')
            fig.savefig(base_name+'.png', dpi=dpi, bbox_inches='tight')
            print('INFO: Saved ranking plot as {}'.format(base_name+'.eps'))
            print('INFO: Saved ranking plot as {}'.format(base_name+'.png'))
            if combine_pdf:
                pp.savefig(fig, bbox_inches='tight')
            else:
                fig.savefig(base_name+'.pdf', dpi=dpi, bbox_inches='tight')
                print('INFO: Saved ranking plot as {}'.format(base_name+'.pdf'))
        if combine_pdf:
            print('INFO: Saved ranking plot as {}'.format(pdf_name))
            pp.close()




from typing import Optional, Union, Dict, List

import pandas as pd
import numpy as np

from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon

from quickstats.plots.template import single_frame, parse_styles, draw_atlas_label

def score_distribution_plot(dfs:Dict[str, pd.DataFrame], hist_options:Dict[str, Dict], 
                            data_options:Optional[Dict[str, Dict]]=None,
                            nbins:int=25, score_name:str='score', weight_name:str='weight',
                            xlabel:str='NN Score', ylabel:str='Fraction of Events / {bin_width}',
                            boundaries:Optional[List]=None, plot_styles:Optional[Dict]=None,
                            analysis_label_options:Optional[Dict]=None):
    styles = parse_styles(plot_styles)
    ax = single_frame(styles=styles, analysis_label_options=analysis_label_options)
    for key in hist_options:
        processes = hist_options[key]['processes']
        hist_style     = hist_options[key].get('style', {})
        combined_df = pd.concat([dfs[process] for process in processes], ignore_index = True)
        norm_weights = combined_df[weight_name]/combined_df[weight_name].sum()
        y, x, _ = ax.hist(combined_df[score_name], nbins, weights=norm_weights, **hist_style,
                          zorder=-5)
    if data_options is not None:
        for key in data_options:
            processes = data_options[key]['processes']
            errorbar_style     = data_options[key].get('style', {})
            combined_df = pd.concat([dfs[process] for process in processes], ignore_index = True)
            norm_weights = combined_df[weight_name]/combined_df[weight_name].sum()
            y, bins = np.histogram(combined_df[score_name], nbins, weights=norm_weights)
            bin_centers  = 0.5*(bins[1:] + bins[:-1])
            ax.errorbar(bin_centers, y, yerr=y**0.5, **errorbar_style)
    ax.yaxis.set_major_locator(MaxNLocator(prune='lower', steps=[10]))
    ax.xaxis.set_major_locator(MaxNLocator(steps=[10]))
    ax.set_xlim(0, 1)
    bin_width = 1/nbins
    ax.set_xlabel(xlabel, **styles['xlabel'])
    ax.set_ylabel(ylabel.format(bin_width=bin_width), **styles['ylabel'])
    handles, labels = ax.get_legend_handles_labels()
    new_handles = [Line2D([], [], c=h.get_edgecolor(), **styles['legend_Line2D'])
                   if isinstance(h, Polygon) else h for h in handles]
    ax.legend(handles=new_handles, labels=labels, **styles['legend'])
    if boundaries is not None:
        for boundary in boundaries:
            ax.axvline(x=boundary, ymin=0, ymax=0.5, linestyle='--', color='k')
    return ax


from typing import Optional, Union, List, Dict

from quickstats.components import AnalysisObject, Likelihood, AsimovGenerator

class AnalysisBase(Likelihood, AsimovGenerator):
    def __init__(self, filename:str, poi_name:str=None, 
                 data_name:str='combData', 
                 config:Optional[Union[Dict, str]]=None,
                 verbosity:Optional[Union[int, str]]="INFO", **kwargs):
        super().__init__(filename=filename,
                         poi_name=poi_name,
                         data_name=data_name,
                         config=config,
                         verbosity=verbosity)


##################################################################################################
# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools
# Author: Alkaid Cheng
# Email: chi.lung.cheng@cern.ch
##################################################################################################
import os
import sys
import copy
import math
import logging
import fnmatch
from typing import List, Optional, Union, Dict, Set

import numpy as np
import pandas as pd

import ROOT

import quickstats
from quickstats import semistaticmethod
from quickstats.components import AbstractObject
from quickstats.components.numerics import is_integer, pretty_value
from quickstats.utils.root_utils import load_macro
from quickstats.utils.roofit_utils import get_variable_attributes, variable_collection_to_dataframe

class ExtendedModel(AbstractObject):
    
    _DEFAULT_NAMES_ = {
        'conditional_globs': 'conditionalGlobs_{mu}',
        'conditional_nuis': 'conditionalNuis_{mu}',
        'nominal_globs': 'nominalGlobs',
        'nominal_nuis': 'nominalNuis',
        'nominal_vars': 'nominalVars',
        'weight': 'weightVar',
        'dataset_args': 'obsAndWeight',
        'asimov': 'asimovData_{mu}',
        'channel_asimov': 'combAsimovData_{label}',
        'nll_snapshot': '{nll_name}_{mu}'
    }
    
    def __init__(self, fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None,
                 data_name:Optional[str]="combData", snapshot_name:Optional[Union[List[str], str]]=None,
                 binned_likelihood:bool=True, tag_as_measurement:Optional[str]=None,
                 fix_cache:bool=True, fix_multi:bool=True, interpolation_code:int=-1,
                 load_extension:bool=True, verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.fname = fname
        self.ws_name = ws_name
        self.mc_name = mc_name
        self.data_name = data_name
        self.initial_snapshots = snapshot_name
        self.binned_likelihood = binned_likelihood
        self.tag_as_measurement = tag_as_measurement
        self.fix_cache = fix_cache
        self.fix_multi = fix_multi
        self.interpolation_code = interpolation_code
        
        if load_extension:
            self.load_extension()
        
        self.initialize()
  
    @property
    def file(self):
        return self._file
    @property
    def workspace(self):
        return self._workspace
    @property
    def model_config(self):
        return self._model_config
    @property
    def pdf(self):
        return self._pdf
    @property
    def data(self):
        return self._data
    @property
    def nuisance_parameters(self):
        return self._nuisance_parameters
    @property
    def global_observables(self):
        return self._global_observables
    @property
    def pois(self):
        return self._pois
    @property
    def observables(self):
        return self._observables
    
    @property
    def initial_snapshots(self):
        return self._initial_snapshots
    
    @initial_snapshots.setter
    def initial_snapshots(self, val):
        if val is None:
            self._initial_snapshots = []
        elif isinstance(val, str):
            self._initial_snapshots = [s.strip() for s in val.split(',') if s]
        elif isinstance(val, list):
            self._initial_snapshots = val
        else:
            raise ValueError("\"initial_snapshots\" must be string or list of strings")
    
    @semistaticmethod
    def load_extension(self):
        for extension in ['RooTwoSidedCBShape', 'FlexibleInterpVarMkII']:
            result = load_macro(extension)
            if (result is not None) and hasattr(ROOT, extension):
                self.stdout.info(f'INFO: Loaded extension module "{extension}"')
            
    @staticmethod
    def load_RooFitObjects():
        load_macro('RooFitObjects')
        
    @semistaticmethod
    def modify_interp_codes(self, ws, interp_code, classes=None):
        if classes is None:
            classes = [ROOT.RooStats.HistFactory.FlexibleInterpVar, ROOT.PiecewiseInterpolation]
        for component in ws.components():
            for cls in classes:
                if (component.IsA() == cls.Class()):
                    component.setAllInterpCodes(interp_code)
                    class_name = cls.Class_Name().split('::')[-1]
                    self.stdout.info('INFO: {} {} interpolation code set to {}'.format(component.GetName(),
                                                                            class_name,
                                                                            interp_code))
        return None
    
    @staticmethod
    def remove_msg_topic(input_arguments=True, numeric_integration=True):
        if input_arguments:
            ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.InputArguments)
        if numeric_integration:
            ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumIntegration)

    @semistaticmethod
    def activate_binned_likelihood(self, ws):
        for component in ws.components():
            try:
                flag = (component.IsA() == ROOT.RooRealSumPdf.Class())
            except:
                flag = (component.ClassName() == "RooRealSumPdf")
            if (flag):
                component.setAttribute('BinnedLikelihood')
                self.stdout.info('INFO: Activated binned likelihood attribute for {}'.format(component.GetName()))
        return None
                          
    @semistaticmethod
    def set_measurement(self, ws, condition):
        for component in ws.components():
            name = component.GetName()
            try:
                flag = (component.IsA() == ROOT.RooAddPdf.Class())
            except:
                flag = (component.ClassName() == "RooAddPdf")
            if flag and condition(name):
                component.setAttribute('MAIN_MEASUREMENT')
                self.stdout.info('INFO: Activated main measurement attribute for {}'.format(name))
        return None
    
    @semistaticmethod
    def deactivate_lv2_const_optimization(self, ws, condition):
        self.stdout.info('INFO: Deactivating level 2 constant term optimization for specified pdfs')
        for component in ws.components():
            name = component.GetName()
            if (component.InheritsFrom(ROOT.RooAbsPdf.Class()) and condition(name)):
                component.setAttribute("NOCacheAndTrack")
                self.stdout.info('INFO: Deactivated level 2 constant term optimization for {}'.format(name))
                
    def load_snapshots(self, snapshot_names:List[str]):
        for snapshot_name in snapshot_names:
            if self.workspace.getSnapshot(snapshot_name):
                self.workspace.loadSnapshot(snapshot_name)
                self.stdout.info(f'INFO: Loaded snapshot "{snapshot_name}"')
            else:
                self.stdout.warning(f'WARNING: Failed to load snapshot "{snapshot_name}"')
            
    def initialize(self):
        if not os.path.exists(self.fname):
            raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
        self.stdout.info('INFO: Opening file "{}"'.format(self.fname))
        file = ROOT.TFile(self.fname) 
        if (not file):
            raise RuntimeError("Something went wrong while loading the root file: {}".format(self.fname))
        # load workspace
        if self.ws_name is None:
            ws_names = [i.GetName() for i in file.GetListOfKeys() if i.GetClassName() == 'RooWorkspace']
            if not ws_names:
                raise RuntimeError("No workspaces found in the root file: {}".format(self.fname))
            if len(ws_names) > 1:
                self.stdout.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
                      " are \"{}\". Will choose the first one by default".format(self.fname, ','.join(ws_names)))
            self.ws_name = ws_names[0]

        ws = file.Get(self.ws_name)
        if not ws:
            raise RuntimeError('failed to load workspace "{}"'.format(self.ws_name))
        self.stdout.info('INFO: Loaded workspace "{}"'.format(self.ws_name))
        # load model config
        if self.mc_name is None:
            mc_names = [i.GetName() for i in ws.allGenericObjects() if 'ModelConfig' in i.ClassName()]
            if not mc_names:
                raise RuntimeError("no ModelConfig object found in the workspace: {}".format(ws_name))
            if len(mc_names) > 1:
                self.stdout.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
                      "Available ModelConfigs are \"{}\". "
                      "Will choose the first one by default".format(ws_name, ','.join(mc_names)))
            self.mc_name = mc_names[0]
        model_config = ws.obj(self.mc_name)
        if not model_config:
            raise RuntimeError('failed to load model config "{}"'.format(self.mc_name))
        self.stdout.info('INFO: Loaded model config "{}"'.format(self.mc_name))
            
        # modify interpolation code
        if self.interpolation_code != -1:
            self.modify_interp_codes(ws, self.interpolation_code,
                                     classes=[ROOT.RooStats.HistFactory.FlexibleInterpVar, ROOT.PiecewiseInterpolation])
        
        # activate binned likelihood
        if self.binned_likelihood:
            self.activate_binned_likelihood(ws)
        
        # set main measurement
        if self.tag_as_measurement:
            self.set_measurement(ws, condition=lambda name: name.startswith(self.tag_as_measurement))
                          
        # deactivate level 2 constant term optimization
            self.deactivate_lv2_const_optimization(ws, 
                condition=lambda name: (name.endswith('_mm') and 'mumu_atlas' in name))

        # load pdf
        pdf = model_config.GetPdf()
        if not pdf:
            raise RuntimeError('Failed to load pdf')
        self.stdout.info('INFO: Loaded model pdf "{}" from model config'.format(pdf.GetName()))
             
        # load dataset
        if self.data_name is None:
            data_names = [i.GetName() for i in ws.allData()]
            if not data_names:
                raise RuntimeError("no datasets found in the workspace: {}".format(ws_name))
            self.data_name = data_names[0]
        data = ws.data(self.data_name)
        if not data:
            raise RuntimeError('Failed to load dataset')
        self.stdout.info('INFO: Loaded dataset "{}" from workspace'.format(data.GetName()))
                
        # load nuisance parameters
        nuisance_parameters = model_config.GetNuisanceParameters()
        if not nuisance_parameters:
            #raise RuntimeError('Failed to load nuisance parameters')
            self.stdout.warning("WARNING: No nuisance parameters found in the workspace. "
                                "An empty set will be loaded by default.")
            nuisance_parameters = ROOT.RooArgSet()
        else:
            self.stdout.info('INFO: Loaded nuisance parameters from model config')
                
        # Load global observables
        global_observables = model_config.GetGlobalObservables()
        if not global_observables:
            self.stdout.warning("WARNING: No nuisance parameters found in the workspace. "
                                "An empty set will be loaded by default.")            
            global_observables = ROOT.RooArgSet()
        else:
            self.stdout.info('INFO: Loaded global observables from model config')                  
    
        # Load POIs
        pois = model_config.GetParametersOfInterest()
        if not pois:
            raise RuntimeError('Failed to load parameters of interest')
        self.stdout.info('INFO: Loaded parameters of interest from model config')
                                  
        # Load observables
        observables = model_config.GetObservables()
        if not observables:
            raise RuntimeError('Failed to load observables')     
        self.stdout.info('INFO: Loaded observables from model config')
        
        self._file                = file
        self._workspace           = ws
        self._model_config        = model_config
        self._pdf                 = pdf
        self._data                = data
        self._nuisance_parameters = nuisance_parameters
        self._global_observables  = global_observables
        self._pois                = pois
        self._observables         = observables
                          
        # Load snapshots
        self.load_snapshots(self.initial_snapshots)
        
        self.remove_msg_topic()
        return None
                
    @semistaticmethod
    def _fix_parameters(self, source:"ROOT.RooArgSet", param_expr=None, param_str='parameter'):
        '''
            source: parameters instance
            param_expr: 
        '''            
        param_dict = self.parse_param_expr(param_expr)
        return self._set_parameters(source, param_dict, mode='fix', param_str=param_str)           
    
    @semistaticmethod
    def _profile_parameters(self, source:"ROOT.RooArgSet", param_expr=None, param_str='parameter'):
        '''
            source: parameters instance
            param_expr: 
        '''                          
        param_dict = self.parse_param_expr(param_expr)
        return self._set_parameters(source, param_dict, mode='free', param_str=param_str)   
    
    def fix_parameters(self, param_expr=None):
        return self._fix_parameters(self.workspace.allVars(), param_expr=param_expr,
                                    param_str='parameter')
    
    def profile_parameters(self, param_expr=None):
        profiled_parameters = self._profile_parameters(self.workspace.allVars(), param_expr=param_expr,
                                                       param_str='parameter') 
        if not profiled_parameters:
            self.stdout.info('Info: No parameters are profiled.')
        return profiled_parameters 
    
    def fix_nuisance_parameters(self, param_expr=None):
        return self._fix_parameters(self.nuisance_parameters, param_expr=param_expr,
                                    param_str='nuisance parameter')
                          
    def fix_parameters_of_interest(self, param_expr=None):
        return self._fix_parameters(self.pois, param_expr=param_expr, param_str='parameter of interest')

    def profile_parameters_of_interest(self, param_expr=None):
        return self._profile_parameters(self.pois, param_expr=param_expr, param_str='parameter of interest')
    
    @semistaticmethod
    def _set_parameters(self, source:"ROOT.RooArgSet", param_dict, mode=None, param_str='parameter'):
        set_parameters = []
        available_parameters = [param.GetName() for param in source]
        for name in param_dict:
            selected_params = [param for param in available_parameters if fnmatch.fnmatch(param, name)]
            if not selected_params:
                self.stdout.warning('WARNING: Parameter "{}" does not exist. No modification will be made.'.format(name))
            for param_name in selected_params:
                self._set_parameter(source[param_name], param_dict[name], mode=mode, param_str=param_str)
                set_parameters.append(source[param_name])

        return set_parameters
    
    @semistaticmethod
    def _set_parameter(self, param, value, mode=None, param_str='parameter'):
        name = param.GetName()
        old_value = param.getVal()
        new_value = old_value
        if isinstance(value, (float, int)):
            new_value = value
        elif isinstance(value, (list, tuple)):
            if len(value) == 3:
                new_value = value[0]
                v_min, v_max = value[1], value[2]
            elif len(value) == 2:
                v_min, v_max = value[0], value[1]
            else:
                raise ValueError('invalid expression for profiling parameter: {}'.format(value))
            # set range
            if (v_min is not None) and (v_max is not None):
                if (new_value < v_min) or (new_value > v_max):
                    new_value = (v_min + v_max)/2
                param.setRange(v_min, v_max)
                self.stdout.info('INFO: Set {} "{}" range to ({},{})'.format(param_str, name, v_min, v_max))
            elif (v_min is not None):
                if (new_value < v_min):
                    new_value = v_min
                # lower bound is zero, if original value is negative, will flip to positive value
                if (v_min == 0) and (old_value < 0):
                    new_value = abs(old_value)
                param.setMin(v_min)
                self.stdout.info('INFO: Set {} "{}" min value to ({},{})'.format(param_str, name, v_min))
            elif (v_max is not None):
                if (new_value > v_max):
                    new_value = v_max
                # upper bound is zero, if original value is positive, will flip to negative value
                if (v_max == 0) and (old_value > 0):
                    new_value = -abs(old_value)                    
                param.setMax(v_max)
                self.stdout.info('INFO: Set {} "{}" max value to ({},{})'.format(param_str, name, v_max))
        if new_value != old_value:
            param.setVal(new_value)              
            self.stdout.info('INFO: Set {} "{}" value to {}'.format(param_str, name, new_value))
        if mode=='fix':
            param.setConstant(1)
            self.stdout.info('INFO: Fixed {} "{}" at value {}'.format(param_str, name, param.getVal()))
        elif mode=='free':
            param.setConstant(0)
            self.stdout.info('INFO: "{}" = [{}, {}]'.format(name, param.getMin(), param.getMax()))
        return None

    @staticmethod
    def set_parameter_defaults(source:"ROOT.RooArgSet", value=None, error=None, constant=None,
                               remove_range=None, target:List[str]=None):

        for param in source:
            if (not target) or (param.GetName() in target):
                if remove_range:
                    param.removeRange()            
                if value is not None:
                    param.setVal(value)
                if error is not None:
                    param.setError(error)
                if constant is not None:
                    param.setConstant(constant)
        return None
    
    @staticmethod
    def parse_param_expr(param_expr):
        param_dict = {}
        # if parameter expression is not empty string or None
        if param_expr: 
            if isinstance(param_expr, str):
                param_dict = ExtendedModel.parse_param_str(param_expr)
            elif isinstance(param_expr, dict):
                param_dict = param_dict
            else:
                raise ValueError('invalid format for parameter expression: {}'.format(param_expr))
        elif param_expr is None:
        # if param_expr is None, all parameters will be parsed as None by default
            param_dict = {param.GetName():None for param in source}
        return param_dict

    @staticmethod
    def parse_param_str(param_str):
        '''
        Example: "param_1,param_2=0.5,param_3=-1,param_4=1,param_5=0:100,param_6=:100,param_7=0:"
        '''
        param_str = param_str.replace(' ', '')
        param_list = param_str.split(',')
        param_dict = {}
        for param_expr in param_list:
            expr = param_expr.split('=')
            # case only parameter name is given
            if len(expr) == 1:
                param_dict[expr[0]] = None
            # case both parameter name and value is given
            elif len(expr) == 2:
                param_name = expr[0]
                param_value = expr[1]
                # range like expression
                if ':' in param_value:
                    param_range = param_value.split(':')
                    if len(param_range) != 2:
                        raise ValueError('invalid parameter range: {}'.format(param_value))
                    param_min = float(param_range[0]) if param_range[0].isnumeric() else None
                    param_max = float(param_range[1]) if param_range[1].isnumeric() else None
                    param_dict[param_name] = [param_min, param_max]
                elif is_integer(param_value):
                    param_dict[param_name] = int(param_value)
                else:
                    param_dict[param_name] = float(param_value)
            else:
                raise ValueError('invalid parameter expression: {}'.format(param))
        return param_dict
    
    @staticmethod
    def randomize_globs(pdf:ROOT.RooAbsPdf, globs:ROOT.RooDataSet, seed:int):
        """Randomize values of global observables (for generating pseudo-experiments)
        """
        # set random seed for reproducible result
        if seed >= 0:
            ROOT.RooRandom.randomGenerator().SetSeed(seed)
        pseudo_globs = pdf.generateSimGlobal(globs, 1)
        pdf_vars = pdf.getVariables()
        pdf_vars.assignValueOnly(pseudo_globs.get(0))
    
    @staticmethod
    def find_unique_prod_components(root_pdf, components, recursion_count=0):
        if (recursion_count > 50):
            raise RuntimeError('find_unique_prod_components detected infinite loop')
        pdf_list = root_pdf.pdfList()
        if pdf_list.getSize() == 1:
            components.add(pdf_list)
            #print('ProdPdf {} is fundamental'.format(pdf_list.at(0).GetName()))
        else:
            for pdf in pdf_list:
                if pdf.ClassName() != 'RooProdPdf':
                    #print('Pdf {} is no RooProdPdf. Adding it.')
                    components.add(pdf)
                    continue
                find_unique_prod_components(pdf, components, recursion_count+1)
                
    @staticmethod 
    def _unfold_constraints(source:ROOT.RooArgSet, obs:ROOT.RooArgSet, nuis:ROOT.RooArgSet, 
                           result:ROOT.RooArgSet=None, recursion:int=0, threshold:int=50):
        if recursion > threshold:
            raise RuntimeError("failed to find unfold constraints: recusion limit exceeded")
        if result is None:
            result = ROOT.RooArgSet()
        for pdf in source:
            class_name = pdf.ClassName()
            if class_name not in ["RooGaussian", "RooLognormal", "RooGamma", "RooPoisson", "RooBifurGauss"]:
                constraint_set = pdf.getAllConstraints(obs.Clone(), nuis.Clone(), ROOT.kFALSE)
                ExtendedModel._unfold_constraints(constraint_set, obs, nuis, result=result, recursion=recursion+1)
            else:
                result.add(pdf)
        return result
    
    def unfold_constraints(self, threshold:int=50):
        source = self.get_all_constraints()
        return self._unfold_constraints(source, self.observables, self.nuisance_parameters, threshold=threshold)
    
    @staticmethod
    def unfold_components(target, source:Set[ROOT.RooAbsArg]):
        result = ROOT.RooArgSet()
        if target in source:
            result.add(target)
        components = None
        if isinstance(target, ROOT.RooAbsPdf):
            components = target.getComponents()
            components.remove(target)
        elif isinstance(target, ROOT.RooProduct):
            components = target.components()
        if (components is not None) and components.getSize():
            for component in components:
                unfolded_components = ExtendedModel.unfold_components(component, source)
                result.add(unfolded_components)
        return result
    
    @semistaticmethod
    def _pair_constraints(self, constraint_set:ROOT.RooArgSet, globs:ROOT.RooArgSet, nuis:ROOT.RooArgSet,
                          base_component:bool=False):
        nuis_list = []
        glob_list = []
        pdf_list = []

        nuis_set = set(nuis)
        for pdf in constraint_set:
            target_np   = None
            target_glob = None
            
            if base_component:
                # getting base pdf component
                components = pdf.getComponents()
                components.remove(pdf)
                if components.getSize():
                    for c1 in components:
                        for c2 in components:
                            if c1 == c2:
                                continue
                            if c2.dependsOn(c1):
                                components.remove(c1)
                    if (components.getSize() > 1):
                        raise RuntimeError("failed to isolate proper nuisance parameter")
                    elif (components.getSize() == 1):
                        target_np = components.first()
                else:
                    for np in nuis:
                        if pdf.dependsOn(np):
                            target_np = np
                            break
            else:
                # getting the actual nuisance parameter
                unfolded_nuis = ExtendedModel.unfold_components(pdf, set(nuis))
                if not unfolded_nuis:
                    for np in nuis:
                        if pdf.dependsOn(np):
                            target_np = np
                            break
                elif len(unfolded_nuis) > 1:
                    raise RuntimeError("failed to isolate proper nuisance parameter")
                else:
                    target_np = unfolded_nuis.first()
                    
            if not target_np:
                self.stdout.warning('WARNING: Could not find nuisance parameter for the constraint: {}'.format(pdf.GetName()))
                continue
            for glob in globs:
                if pdf.dependsOn(glob):
                    target_glob = glob
                    break
            if not target_glob:
                self.stdout.warning('WARNING: Could not find global observable for the constraint: {}'.format(pdf.GetName()))
                continue
            nuis_list.append(target_np)
            glob_list.append(target_glob)
            pdf_list.append(pdf)
        return nuis_list, glob_list, pdf_list
    
    def pair_nuis_and_glob_obs(self):
        constraint_set = self.unfold_constraints()
        nuis_list, glob_list, _ = self._pair_constraints(constraint_set, self.global_observables, self.nuisance_parameters)
        return nuis_list, glob_list

    def pair_constraints(self, to_str=False, base_component:bool=False):
        constraint_set = self.unfold_constraints()
        nuis_list, glob_list, pdf_list = self._pair_constraints(constraint_set, self.global_observables, 
                                                                self.nuisance_parameters,
                                                                base_component=base_component)
        if to_str:
            nuis_names = [i.GetName() for i in nuis_list]
            glob_names = [i.GetName() for i in glob_list]
            pdf_names  = [i.GetName() for i in pdf_list]
            size = len(nuis_names)
            return [(pdf_names[i], nuis_names[i], glob_names[i]) for i in range(size)]
        return nuis_list, glob_list, pdf_list
    
    def get_constrained_nuisance_parameters(self):
        constrained_nuis, _, _ = self.pair_constraints()
        return constrained_nuis
    
    def get_unconstrained_nuisance_parameters(self, constrained_nuis=None):
        # do not re-isolate constrained nuisance parameters
        if constrained_nuis is None:
            constrained_nuis = self.get_constrained_nuisance_parameters()
        unconstrained_nuis = list(set(self.nuisance_parameters) - set(constrained_nuis))
        return unconstrained_nuis
    
    def set_constrained_nuisance_parameters_to_nominal(self, set_constant:bool=False):
        constrained_nuis, _, constrain_pdf = self.pair_constraints()
        for nuis, pdf in zip(constrained_nuis, constrain_pdf):
            pdf_class = pdf.ClassName()
            if pdf_class in ["RooGaussian", "RooBifurGauss"]:
                nuis.setVal(0)
            elif pdf_class == "RooPoisson":
                nuis.setVal(1)
            else:
                raise RuntimeError(f"constraint term \"{pdf.GetName()}\" has unsupported type \"{pdf_class}\"")
        if set_constant:
            for nuis in constrained_nuis:
                nuis.setConstant(1)
    
    def get_all_constraints(self):
        all_constraints = ROOT.RooArgSet()
        cache_name = "CACHE_CONSTR_OF_PDF_{}_FOR_OBS_{}".format(self.pdf.GetName(), 
                     ROOT.RooNameSet(self.data.get()).content())                 
        constr = self.workspace.set(cache_name)
        if constr:
            # retrieve constrains from cache     
            all_constraints.add(constr)
        else:
            # load information needed to determine attributes from ModelConfig 
            obs = self.observables.Clone()
            nuis = self.nuisance_parameters.Clone()
            all_constraints = self.pdf.getAllConstraints(obs, nuis, ROOT.kFALSE)
            
        # take care of the case where we have a product of constraint terms
        temp_all_constraints = ROOT.RooArgSet(all_constraints.GetName())
        for constraint in all_constraints:
            if constraint.IsA() == ROOT.RooProdPdf.Class():
                buffer = ROOT.RooArgSet()
                ExtendedModel.find_unique_prod_components(constraint, buffer)
                temp_all_constraints.add(buffer)
            else:
                temp_all_constraints.add(constraint)
        return temp_all_constraints
    
    def inspect_constrained_nuisance_parameter(self, nuis, constraints):
        nuis_name = nuis.GetName()
        self.stdout.info('INFO: On nuisance parameter {}'.format(nuis_name))
        nuip_nom = 0.0
        prefit_variation = 1.0
        found_constraint = ROOT.kFALSE
        found_gaussian_constraint = ROOT.kFALSE
        constraint_type = None
        for constraint in constraints:
            constr_name = constraint.GetName()
            if constraint.dependsOn(nuis):
                found_constraint = ROOT.kTRUE
                constraint_type = 'unknown'
                # Loop over global observables to match nuisance parameter and
                # global observable in case of a constrained nuisance parameter
                found_global_observable = ROOT.kFALSE
                for glob_obs in self.global_observables:
                    if constraint.dependsOn(glob_obs):
                        found_global_observable = ROOT.kTRUE
                        # find constraint width in case of a Gaussian
                        if constraint.IsA() == ROOT.RooGaussian.Class():
                            found_gaussian_constraint = ROOT.kTRUE
                            constraint_type = 'gaus'
                            old_sigma_value = 1.0
                            found_sigma = ROOT.kFALSE
                            for server in constraint.servers():
                                if (server != glob_obs) and (server != nuis):
                                    old_sigma_value = server.getVal()
                                    found_sigma = ROOT.kTRUE
                            if math.isclose(old_sigma_value, 1.0, abs_tol=0.001):
                                old_sigma_value = 1.0
                            if not found_sigma:
                                self.stdout.info('INFO: Sigma for pdf {} not found. Uisng 1.0.'.format(constr_name))
                            else:
                                self.stdout.info('INFO: Uisng {} for sigma of pdf {}'.format(old_sigma_value, constr_name))

                            prefit_variation = old_sigma_value
                        elif constraint.IsA() == ROOT.RooPoisson.Class():
                            constraint_type = 'pois'
                            tau = glob_obs.getVal()
                            self.stdout.info('INFO: Found tau {} of pdf'.format(constr_name))
                            prefit_variation = 1. / math.sqrt(tau)
                            self.stdout.info('INFO: Prefit variation is {}'.format(prefit_variation))
                            nuip_nom = 1.0
                            self.stdout.info("INFO: Assume that {} is nominal value of the nuisance parameter".format(nuip_nom))
        return prefit_variation, constraint_type, nuip_nom
        
    def set_initial_errors(self, source:Optional["ROOT.RooArgSet"]=None):
        if not source:
            source = self.nuisance_parameters
    
        all_constraints = self.get_all_constraints()
        for nuis in source:
            nuis_name = nuis.GetName()
            prefit_variation, constraint_type, _ = self.inspect_constrained_nuisance_parameter(nuis, all_constraints)
            if constraint_type=='gaus':
                self.stdout.info('INFO: Changing error of {} from {} to {}'.format(nuis_name, nuis.getError(), prefit_variation))
                nuis.setError(prefit_variation)
                nuis.removeRange()    
        return None
    
    def get_poi(self, poi_name:Optional[str]=None, strict:bool=False):
        if poi_name is None:
            poi = self.pois.first()
            self.stdout.info(f"INFO: POI name not specified. The first POI \"{poi.GetName()}\" is used by default.")
        else:
            poi = self.workspace.var(poi_name)
        if not poi:
            raise RuntimeError(f"workspace does not contain the variable \"{poi_name}\"")
        if strict and (poi not in list(self.pois)):
            raise RuntimeError(f"workspace variable \"{poi_name}\" is not part of the POIs")
        return poi
    
    def _load_obs_and_weight(self, obs_and_weight:Optional[Union["ROOT.RooArgSet",str]]=None, 
                             weight_var:Optional[Union["ROOT.RooRealVar",str]]=None):
        # get the weight variable
        if weight_var is None:
            weight_name = self._DEFAULT_NAMES_['weight']
            weight_var = self.workspace.var(weight_name)
            if not weight_var:
                weight_var = ROOT.RooRealVar(weight_name, weight_name, 1)
                getattr(self.workspace, "import")(weight_var)
        elif isinstance(weight_var, str):
            weight_var = self.workspace.var(weight_var)
            if not weight_var:
                raise RuntimeError('weight variable "{}" not found in workspace'.format(weight_var))
        elif not isinstance(weight_var, ROOT.RooRealVar):
            raise ValueError('weight variable must be of RooRealVar type')
                
        # get the obs_and_weight arg set
        if obs_and_weight is None:
            default_name = self._DEFAULT_NAMES_['dataset_args']
            obs_and_weight = self.workspace.set(default_name)
            if not obs_and_weight:
                obs_and_weight = ROOT.RooArgSet()
                obs_and_weight.add(self.observables)            
                obs_and_weight.add(weight_var)
                self.workspace.defineSet(default_name, obs_and_weight)
        elif isinstance(obs_and_weight, str):
            obs_and_weight = self.workspace.set(obs_and_weight)
            if not obs_and_weight:
                raise RuntimeError('named set "{}" not found in workspace'.format(obs_and_weight))
        elif not isinstance(obs_and_weight, ROOT.RooArgSet):
            raise ValueError('the argument "obs_and_weight" must be of RooArgSet type')
        return obs_and_weight, weight_var
    
    @staticmethod
    def get_object_map(object_dict:Dict, object_name:str):
        ExtendedModel.load_RooFitObjects()
        if object_name not in ["RooDataSet", "RooAbsPdf"]:
            raise ValueError("unsupported object \"{}\"".format(object_name))
        object_map = ROOT.std.map(f"string, {object_name}*")()
        object_map.keepalive = list()
        for c, d in object_dict.items():
            object_map.keepalive.append(d)
            object_map.insert(object_map.begin(), ROOT.std.pair(f"const string, {object_name}*")(c, d))
        return object_map
    
    @staticmethod
    def get_dataset_map(dataset_dict:Dict):
        ExtendedModel.load_RooFitObjects()
        dsmap = ROOT.std.map('string, RooDataSet*')()
        dsmap.keepalive = list()
        for c, d in dataset_dict.items():
            dsmap.keepalive.append(d)
            dsmap.insert(dsmap.begin(), ROOT.std.pair("const string, RooDataSet*")(c, d))
        return dsmap
    
    @staticmethod
    def get_pdf_map(pdf_dict:Dict):
        ExtendedModel.load_RooFitObjects()
        pdfmap = ROOT.std.map('string, RooAbsPdf*')()
        pdfmap.keepalive = list()
        for c, d in pdf_dict.items():
            pdfmap.keepalive.append(d)
            pdfmap.insert(pdfmap.begin(), ROOT.std.pair("const string, RooAbsPdf*")(c, d))
        return pdfmap
    
    def generate_asimov_from_pdf(self, name:str="asimovData", pdf:Optional["ROOT.RooAbsPdf"]=None,
                                 obs_and_weight:Optional[Union["ROOT.RooArgSet",str]]=None, 
                                 weight_var:Optional[Union["ROOT.RooRealVar",str]]=None,
                                 extra_args=None):
        
        pdf = pdf if pdf is not None else self.pdf
        if isinstance(pdf, ROOT.RooSimultaneous):
            raise ValueError("this method should not be called from a simultaneous pdf")
        
        obs_and_weight, weight_var = self._load_obs_and_weight(obs_and_weight, weight_var)
        
        # get the combined arg set for the asimov dataset
        arg_set = ROOT.RooArgSet()
        arg_set.add(obs_and_weight)
        if extra_args is not None:
            if isinstance(extra_args, list):
                for arg in extra_args:
                    arg_set.add(arg)
            else:
                arg_set.add(extra_args)
                
        asimov_data = ROOT.RooDataSet(name, name, arg_set, ROOT.RooFit.WeightVar(weight_var))
        
        # generate observables defined by the pdf associated with this state
        obs = pdf.getObservables(self.observables)
        target_obs = obs.first()
        expected_events = pdf.expectedEvents(obs)
        #print("INFO: Generating Asimov for pdf {}".format(pdf.GetName()))
        for i in range(target_obs.numBins()):
            target_obs.setBin(i)
            norm = pdf.getVal(obs)*target_obs.getBinWidth(i)
            n_events = norm*expected_events
            if n_events <= 0:
                self.stdout.warning("WARNING: Detected bin with zero expected events ({})! Please check"
                      "your inputs. Obs = {}, bin = {}".format(n_events, target_obs.GetName(), i))
            elif (n_events > 0) and (n_events < 1e18):
                self.stdout.debug("pdf={}, obs={}, bin={}, val={}".format(
                    pdf.GetName(), target_obs.GetName(), i, n_events))
                asimov_data.add(self.observables, n_events)
            else:
                raise RuntimeError(f"detected pdf bin with nan (pdf={pdf.GetName()},obs={target_obs.GetName()},bin={i})")

        if (asimov_data.sumEntries() != asimov_data.sumEntries()):
            raise RuntimeError("asimov data sum entries is nan")
        return asimov_data
    
    def match_globs(self):
        nuis_list, globs_list, pdf_list = self.pair_constraints(base_component=True)
        for np, glob in zip(nuis_list, globs_list):
            glob.setVal(np.getVal())
            self.stdout.debug("set glob {} to val {}".format(glob.GetName(), glob.getVal()))
        """
        for nuis, globs, pdf in zip(nuis_list, globs_list, pdf_list):
            pdf_class = pdf.ClassName()
            if pdf_class in ["RooGaussian", "RooBifurGauss"]:
                globs.setVal(nuis.getVal())
            elif pdf_class == "RooPoisson":
                from pdb import set_trace
                set_trace()                
                globs.setVal(nuis.getVal()*globs.getVal())
            else:
                raise RuntimeError(f"constraint term \"{pdf.GetName()}\" has unsupported type \"{pdf_class}\"")
            self.stdout.debug("set glob {} to val {}".format(globs.GetName(), globs.getVal()))
        """

    def generate_asimov(self, poi_name:str, poi_val:Optional[float]=None, 
                        poi_profile:Optional[float]=None,
                        do_fit:bool=True,
                        modify_globs:bool=True,
                        do_import:bool=True,
                        asimov_name:Optional[str]=None,
                        asimov_snapshot:Optional[str]=None,
                        channel_asimov_name:Optional[str]=None,
                        dataset:Optional["ROOT.RooDataSet"]=None,
                        constraint_option:int=0,
                        restore_states:int=0,
                        minimizer_options:Optional[Dict]=None, 
                        nll_options:Optional[Union[Dict, List]]=None,
                        snapshot_names:Optional[Dict]=None):
        """
            Generate Asimov dataset.
            
            Note:
                Nominal (initial values) snapshots of nuisance parameters and global are saved
                as "nominalNuis" and "nominalGlobs" if not already exist. Conditional snapshots
                are saved as "conditionalNuis_{mu}" and "conditionalGlobs_{mu}" irrespective of
                whether nuisance parameter profiling is performed and whether conditional mle
                is used for the profiling. This is to faciliate the use case of Asymptotic limit
                calculation. The names of these snahpshots can be customized via the
                `snapshot_names` option.
        
            Arguments:
                poi_name: str
                    Name of the parameter of interest (POI).
                poi_val: (Optional) float
                    Generate asimov data with POI set at the specified value. If None, POI will be kept
                    at the post-fit value if a fitting is performed or the pre-fit value if no fitting
                    is performed.
                poi_profile: (Optional) float
                    Perform nuisance parameter profiling with POI set at the specified value. This option
                    is only effective if do_fit is set to True. If None, POI is set floating 
                    (i.e. unconditional maximum likelihood estimate). 
                do_fit: bool, default=True    
                    Perform nuisance parameter profiling with a fit to the given dataset.
                modify_globs: bool, default=True
                    Match the values of nuisance parameters and the corresponding global observables when
                    generating the asimov data. This is important for making sure the asimov data has the 
                    (conditional) minimal NLL.
                constraint_option: int, default=0
                    Customize the target of nuisance paramaters involved in the profiling.
                    Case 0: All nuisance parameters are allowed to float;
                    Case 1: Constrained nuisance parameters are fixed to 0. Unconstrained nuisrance
                            parameters are allowed to float.
                restore_states: int, default=0
                    Restore variable states at the end of asimov data generation.
                    Case 0: All variable states will be restored.
                    Case 1: Only global observable states will be restored.
                do_import: bool, default=True
                    Import the generated asimov data to the current workspace.
                asimov_name: (Optional) str
                    Name of the generated asimov dataset. If None, defaults to "asimovData_{mu}" where
                    `{mu}` will be replaced by the value of `poi_val`. Other keywords are: `{mu_cond}`
                    which is the the value of `poi_profile`.
                asimov_snapshot: (Optional) str
                    Name of the snapshot taken right after asimov data generation. If None, no snapshot
                    will be saved.
                channel_asimov_name: (Optional) str
                    Name of the asimov dataset in each category of a simultaneous pdf. If None,
                    defaults to "combAsimovData_{label}" where `{label}` will be replaced by the
                dataset: (Optional) ROOT.RooDataSet
                    Dataset based on which the negative log likelihood (NLL) is created for nuisance parameter
                    profiling. If None, default to self.data.
                minimizer_options: (Optional) dict
                    Options for minimization during nuisance parameter profiling. If None, defaults to
                    ExtendedMinimizer._DEFAULT_MINIMIZER_OPTION_
                nll_options: (Optional) dict, list
                    Options for NLL creation during nuisance parameter profiling. If None, defaults to
                    ExtendedMinimizer._DEFAULT_NLL_OPTION_
                snapshot_names: (Optional) dict
                    A dictionary containing a map of the snapshot type and the snapshot names. The default
                    namings are stored in ExtendedModel._DEFAULT_NAMES_.
        """
        # define simplified ws variable names
        ws = self.workspace
        all_globs = self.global_observables
        all_nuis  = self.nuisance_parameters
        
        # define names used for various objects
        names = copy.deepcopy(self._DEFAULT_NAMES_)
        if snapshot_names is not None:
            names.update(snapshot_names)
        if asimov_name is not None:
            names['asimov'] = asimov_name
        if channel_asimov_name is not None:
            names['channel_asimov'] = channel_asimov_name
        nom_vars_name = names['nominal_vars']
        nom_glob_name = names['nominal_globs']
        nom_nuis_name = names['nominal_nuis']
        con_glob_name = names['conditional_globs']
        con_nuis_name = names['conditional_nuis']

        poi = self.get_poi(poi_name)
        
        # take snapshot of initial states of all variables
        ws.saveSnapshot(nom_vars_name, self.workspace.allVars())
        if not ws.getSnapshot(nom_glob_name):
            self.stdout.info('INFO: Saving snapshot "{}"'.format(nom_glob_name))
            ws.saveSnapshot(nom_glob_name, all_globs)
        if not ws.getSnapshot(nom_nuis_name):
            self.stdout.info('INFO: Saving snapshot "{}"'.format(nom_nuis_name))
            ws.saveSnapshot(nom_nuis_name, all_nuis)
        
        if do_fit:
            from quickstats.components import ExtendedMinimizer
            if dataset is None:
                dataset = self.data
            minimizer = ExtendedMinimizer("Minimizer", self.pdf, dataset)
            if minimizer_options is None:
                minimizer_options = copy.deepcopy(minimizer._DEFAULT_MINIMIZER_OPTION_)
            if nll_options is None:
                nll_options = copy.deepcopy(minimizer._DEFAULT_NLL_OPTION_)
            minimizer.configure(**minimizer_options)
            if constraint_option == 0:
                pass
            elif constraint_option == 1:
                self.set_constrained_nuisance_parameters_to_nominal()
            else:
                raise ValueError(f"unsupported constraint option: {constraint_option}")
            if isinstance(nll_options, dict):
                minimizer.configure_nll(all_nuis, all_globs, **nll_options)
            elif isinstance(nll_options, list):
                minimizer.nll_command_list = nll_options
            else:
                raise ValueError(f"unsupported nll options format")
            if poi_profile is not None:
                # conditional mle
                poi.setVal(poi_profile)
                poi.setConstant(1)
            else:
                # unconditional mle
                poi.setConstant(0)
            minimizer.minimize()

        if poi_val is not None:
            poi.setVal(poi_val)
        poi.setConstant(0)
        poi_val = poi.getVal()
        
        # match values of global observables to the corresponding NPs
        if modify_globs:
            self.match_globs()

        if do_fit:
            ws.saveSnapshot(con_glob_name.format(mu=pretty_value(poi_profile)), all_globs)
            ws.saveSnapshot(con_nuis_name.format(mu=pretty_value(poi_profile)), all_nuis)
        else:
            ws.saveSnapshot(con_glob_name.format(mu=pretty_value(poi_val)), all_globs)
            ws.saveSnapshot(con_nuis_name.format(mu=pretty_value(poi_val)), all_nuis)

        asimov_data_name = names['asimov'].format(mu=pretty_value(poi_val), 
                                                  mu_cond=pretty_value(poi_profile))
        channel_asimov_data_name = names['channel_asimov']
        sim_pdf = self.pdf
        if not isinstance(sim_pdf, ROOT.RooSimultaneous):
            asimov_data = self.generate_asimov_from_pdf(asimov_data_name, sim_pdf)
        else:
            asimov_data_map = {}
            channel_cat = sim_pdf.indexCat()
            n_cat = len(channel_cat)
            for i in range(n_cat):
                channel_cat.setIndex(i)
                label = channel_cat.getLabel()
                pdf_cat = sim_pdf.getPdf(label)
                name = channel_asimov_data_name.format(index=i, label=label)
                asimov_data_map[label] = self.generate_asimov_from_pdf(name, pdf_cat, extra_args=channel_cat)

            obs_and_weight, weight_var = self._load_obs_and_weight()
            dataset_map = ExtendedModel.get_dataset_map(asimov_data_map)
            asimov_data = ROOT.RooDataSet(asimov_data_name, asimov_data_name, 
                                          ROOT.RooArgSet(obs_and_weight, channel_cat),
                                          ROOT.RooFit.Index(channel_cat),
                                          ROOT.RooFit.Import(dataset_map),
                                          ROOT.RooFit.WeightVar(weight_var))
        if do_import:
            getattr(ws, "import")(asimov_data)
            self.stdout.info(f'INFO: Generated Asimov Dataset \"{asimov_data_name}\"')
            
        if asimov_snapshot is not None:
            self.stdout.info('INFO: Saving snapshot "{}"'.format(asimov_snapshot))
            ws.saveSnapshot(asimov_snapshot, self.workspace.allVars())
    
        if restore_states == 0:
            # load back a snapshot of all varaible's initial states
            ws.loadSnapshot(nom_vars_name)
        elif restore_states == 1:
            # load back a snapshot of the initial global observable states
            ws.loadSnapshot(nom_glob_name)
        elif restore_states == 2:
            pass
        else:
            raise ValueError(f"unsupported restore state option \"{restore_states}\"")
        
        return asimov_data
    
    def generate_toys(self, n_toys:int=1, seed:int=0, binned:bool=True, 
                      do_import:bool=True, name="toyData_{index}_seed_{seed}"):
        ws = self.workspace
        # take snapshot of initial states of all variables
        self.workspace.saveSnapshot("tmp", self.workspace.allVars())
        
        if binned:
            if self.pdf.ClassName() == "RooSimultaneous":
                index_cat = self.pdf.indexCat()
                for cat in index_cat:
                    pdf_i = self.pdf.getPdf(cat.first)
                    pdf_i.setAttribute("GenerateToys::Binned")
            else:
                self.pdf.setAttribute("GenerateToys::Binned")
        
        toys = []
        args = [self.observables, ROOT.RooFit.Extended(), ROOT.RooFit.AutoBinned(True)]
        if binned:
            args.append(ROOT.RooFit.GenBinned("GenerateToys::Binned"))
        
        for i in range(n_toys):
            self.randomize_globs(self.pdf, self.global_observables, seed)
            toy = self.pdf.generate(*args)
            toy_name = name.format(seed=seed, index=i)
            toy.SetName(toy_name)
            if do_import:
                getattr(ws, "import")(toy)
                self.stdout.info(f'INFO: Generated toy dataset "{toy_name}"')
            toys.append(toy)
        
        ws.loadSnapshot("tmp")
        
        return toys
    
    def save(self, fname:str, recreate:bool=True):
        self.workspace.writeToFile(fname, recreate)
    
    @semistaticmethod
    def load_ws(self, fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
        if not os.path.exists(fname):
            raise FileNotFoundError('workspace file {} does not exist'.format(fname))
        file = ROOT.TFile(fname)
        if (not file):
            raise RuntimeError("Something went wrong while loading the root file: {}".format(fname))        
        # load workspace
        if ws_name is None:
            ws_names = [i.GetName() for i in file.GetListOfKeys() if i.GetClassName() == 'RooWorkspace']
            if not ws_names:
                raise RuntimeError("No workspaces found in the root file: {}".format(fname))
            if len(ws_names) > 1:
                self.stdout.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
                      " are \"{}\". Will choose the first one by default".format(fname, ','.join(ws_names)))
            ws_name = ws_names[0]
        ws = file.Get(ws_name)
        if not ws:
            raise RuntimeError('Failed to load workspace: "{}"'.format(ws_name))
        # load model config
        if mc_name is None:
            mc_names = [i.GetName() for i in ws.allGenericObjects() if 'ModelConfig' in i.ClassName()]
            if not mc_names:
                raise RuntimeError("no ModelConfig object found in the workspace: {}".format(ws_name))
            if len(mc_names) > 1:
                self.stdout.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
                      "Available ModelConfigs are \"{}\". "
                      "Will choose the first one by default".format(ws_name, ','.join(mc_names)))
            mc_name = mc_names[0]     
        mc = ws.obj(mc_name)
        if not mc:
            raise RuntimeError('Failed to load model config "{}"'.format(mc_name))
        return file, ws, mc
    
    @staticmethod
    def get_nuisance_parameter_names(fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
        ExtendedModel.load_extension()
        file, ws, mc = ExtendedModel.load_ws(fname, ws_name, mc_name)
        nuisance_parameters = mc.GetNuisanceParameters()
        if not nuisance_parameters:
            raise RuntimeError('Failed to load nuisance parameters')
        return [nuis.GetName() for nuis in nuisance_parameters]
    
    @staticmethod
    def get_poi_names(fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
        ExtendedModel.load_extension()
        file, ws, mc = ExtendedModel.load_ws(fname, ws_name, mc_name)
        pois = mc.GetParametersOfInterest()
        if not pois:
            raise RuntimeError('Failed to load parameters of interest')        
        return [poi.GetName() for poi in pois]

    @staticmethod
    def get_dataset_names(fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
        ExtendedModel.load_extension()
        file, ws, mc = ExtendedModel.load_ws(fname, ws_name, mc_name)
        datasets = ws.allData()
        if not datasets:
            raise RuntimeError('Failed to load datasets')
        return [dataset.GetName() for dataset in datasets]
    
    def get_category_map(self, pdf=None):
        if pdf is None:
            pdf = self.pdf
        if not isinstance(pdf, ROOT.RooSimultaneous):
            raise ValueError("input pdf is not a simultaneous pdf")            
        category_map = {}
        index_cat = pdf.indexCat()
        n_cat = len(index_cat)
        for i in range(n_cat):
            index_cat.setIndex(i)
            label = index_cat.getLabel()
            pdf_cat = pdf.getPdf(label)
            obs = pdf_cat.getObservables(self.observables)
            target_obs = obs.first()
            category_map[label] = {}
            category_map[label]['index'] = i
            category_map[label]['pdf'] = pdf_cat.GetName()
            category_map[label]['observable'] = target_obs.GetName()
        return category_map
    
    @staticmethod
    def get_dataset_values(dataset):
        parameters = dataset.get()
        observables = []
        category = None
        for p in parameters:
            if isinstance(p, ROOT.RooRealVar):
                observables.append(p)
            elif isinstance(p, ROOT.RooCategory):
                if category is None:
                    category = p
                else:
                    raise RuntimeError("multiple RooCategory instances found in the dataset")
            else:
                raise RuntimeError(f"unknown object type \"{type(p)}\" found in the dataset")
        n_entries = dataset.numEntries()
        result = {}
        """
        # only available after ROOT 6.24
        data_buffer = ROOT.RooBatchCompute.RunContext()
        dataset.getBatches(data_buffer, 0, n_entries)
        for obs in observables:
            obs_name = obs.GetName()
            data = np.asarray(data_buffer.getBatch(obs))
            result[obs_name] = data
        """
        for obs in observables:
            obs_name = obs.GetName()
            result[obs_name] = np.empty(n_entries)
            for i in range(n_entries):
                dataset.get(i)
                result[obs_name][i] = obs.getVal()
        result['weight'] = np.asarray(dataset.getWeightBatch(0, n_entries))
        if category is not None:
            result['label'] = np.empty(n_entries, dtype=object)
            result['index'] = np.empty(n_entries)
            for i in range(n_entries):
                dataset.get(i)
                result['label'][i] = category.getLabel()
                result['index'][i] = category.getIndex()
        return result
    
    def get_dataset_distributions(self, dataset):
        distributions = {}
        dataset_values = self.get_dataset_values(dataset)
        if 'label' not in dataset_values:
            raise RuntimeError("no categories defined in the dataset")
        category_map = self.get_category_map()
        for cat in category_map:
            distributions[cat] = {}
            observable = category_map[cat]['observable']
            if observable not in dataset_values:
                raise RuntimeError(f"no data associated with the observable \"{observable}\" found in the dataset")
            distributions[cat]['observable'] = observable
            mask = (dataset_values['label'] == cat)
            x = dataset_values[observable][mask]
            ind = np.argsort(x)
            x = x[ind]
            y = dataset_values['weight'][mask][ind]
            distributions[cat]['x'] = x
            distributions[cat]['y'] = y
        return distributions
    
    @semistaticmethod
    def get_simul_pdf_distributions(self, pdf, observables):
        if not isinstance(pdf, ROOT.RooSimultaneous):
            raise ValueError("input pdf is not a simultaneous pdf")
        distributions = {}
        channel_cat = pdf.indexCat()
        n_cat = len(channel_cat)
        for i in range(n_cat):
            channel_cat.setIndex(i)
            label = channel_cat.getLabel()
            pdf_cat = pdf.getPdf(label)
            obs = pdf_cat.getObservables(observables)
            target_obs = obs.first()
            expected_events = pdf_cat.expectedEvents(obs)
            n_bins = target_obs.numBins()
            obs_name = target_obs.GetName()
            distributions[label] = {'x':[], 'y':[], 'observable': obs_name}
            for i in range(n_bins):
                target_obs.setBin(i)
                norm = pdf_cat.getVal(obs)*target_obs.getBinWidth(i)
                n_events = norm*expected_events
                bin_value = target_obs.getVal()
                if n_events <= 0:
                    self.stdout.warning("WARNING: Detected bin with zero expected events ({})! "
                                    "Please check your inputs. Obs = {}, bin = {}".format(
                                    n_events, target_obs.GetName(), i))
                elif (n_events > 0) and (n_events < 1e18):
                    distributions[label]['x'].append(bin_value)
                    distributions[label]['y'].append(n_events)
                else:
                    raise RuntimeError(f"detected pdf bin with nan (pdf={pdf.GetName()},obs={target_obs.GetName()},bin={i})")
        return distributions
    
    def get_distributions(self):
        return self.get_simul_pdf_distributions(self.pdf, self.observables)
    
    def get_categories(self):
        return list(self.get_category_map())
    
    def get_collected_distributions(self, current_distributions=True,
                                    datasets:Optional[Union[List[str],List[ROOT.RooDataSet]]]=None,
                                    snapshots:Optional[List[str]]=None):
        collected_distributions = {}
        
        if current_distributions:
            distributions = self.get_distributions()
            for category in distributions:
                collected_distributions[category] = {"Current": distributions[category]}
                
        if datasets is not None:
            for dataset in datasets:
                if isinstance(dataset, str):
                    dataset_name = dataset
                    dataset = self.workspace.data(dataset)
                    if not dataset:
                        raise ValueError(f"dataset \"{dataset_name}\" does not exist")
                dataset_name = dataset.GetName()
                distributions = self.get_dataset_distributions(dataset)
                for category in distributions:
                    if category not in collected_distributions:
                        collected_distributions[category] = {}
                    collected_distributions[category][dataset_name] = distributions[category]
        if snapshots is not None:
            self.workspace.saveSnapshot("tmp", self.workspace.allVars())
            try:
                for snapshot in snapshots:
                    exist = self.workspace.loadSnapshot(snapshot)
                    if not exist:
                        raise RuntimeError("snapshot \"{}\" not found in workspace".format(snapshot))
                    distributions = self.get_distributions()
                    for category in distributions:
                        if category not in collected_distributions:
                            collected_distributions[category] = {}
                        collected_distributions[category][snapshot] = distributions[category]
            finally:
                self.workspace.loadSnapshot("tmp")
        return collected_distributions
            
    def plot_distributions(self, category:str, current_distributions=True, 
                           datasets:Optional[Union[List[str],List[ROOT.RooDataSet]]]=None,
                           snapshots:Optional[Union[List[str], List[ROOT.RooArgSet]]]=None):
        from quickstats.plots.template import plot_distributions
        category_map = self.get_category_map()
        if category not in category_map:
            raise ValueError("category \"{}\" not found in workspace".format(category))
        collected_distributions = self.get_collected_distributions(current_distributions,
                                                                   datasets, snapshots=snapshots)
        xlabel = category_map[category]['observable']
        ylabel = "Events"
        return plot_distributions(collected_distributions[category], xlabel=xlabel, ylabel=ylabel)
    
    def as_dataframe(self, variable_type:str):
        if variable_type == "pois":
            vcollection = self.pois
        elif variable_type == "nuisance_parameters":
            vcollection = self.nuisance_parameters
        elif variable_type == "global_observables":
            vcollection = self.global_observables
        elif variable_type == "constrained_nuisance_parameters":
            vcollection = self.get_constrained_nuisance_parameters()
        elif variable_type == "unconstrained_nuisance_parameters":
            vcollection = self.get_unconstrained_nuisance_parameters()
        elif variable_type == "constraints":
            data = self.pair_constraints(to_str=True)
            df = pd.DataFrame(data, columns =['pdf', 'nuisance parameters', 'global observables'])
            return df
        else:
            raise ValueError("unknown variable type")
        df = variable_collection_to_dataframe(vcollection)
        return df    
    
    def print_summary(self, items:Optional[List]=None, suppress_print:bool=False, detailed:bool=True,
                      include_patterns:Optional[List]=None, exclude_patterns:Optional[List]=None,
                      save_as:Optional[str]=None):
        if items is None:
            items = ['workspace', 'dataset', 'category', 'observable', 
                     'poi', 'detailed_nuisance_parameter']
        summary_str = ""
        # workspace
        if 'workspace' in items:
            summary_str += f"Workspace:\n"
            summary_str += f"\t{self.workspace.GetName()}\n"
        # datasets
        if 'dataset' in items:
            datasets = self.workspace.allData()
            summary_str += f"Datasets ({len(datasets)}):\n"
            summary_str += "".join([f"\t{obs.GetName()}\n" for obs in datasets])
        # categories
        if 'category' in items:
            if isinstance(self.pdf, ROOT.RooSimultaneous):
                channel_cat = self.pdf.indexCat()
                n_cat = len(channel_cat)
                summary_str += f"Categories ({n_cat}):\n"
                for i in range(n_cat):
                    channel_cat.setIndex(i)
                    label = channel_cat.getLabel()
                    summary_str += f"\t{label}\n"
        # observables
        if 'observable' in items:
            summary_str += f"Observables ({len(self.observables)}):\n"
            summary_str += "".join([f"\t{obs.GetName()}\n" for obs in self.observables])
        # pois, NPs
        param_strs     = []
        param_sets     = []
        if 'poi' in items:
            pois = self.pois            
            param_strs.append('POIs')
            param_sets.append(pois)
        if 'detailed_nuisance_parameter' in items:
            constrained_nps = self.get_constrained_nuisance_parameters()
            unconstrained_nps = self.get_unconstrained_nuisance_parameters(constrained_nps)            
            param_strs += ['Constrained NPs', 'Unconstrained NPs']
            param_sets += [constrained_nps, unconstrained_nps]
        elif 'nuisance_parameter' in items:
            param_strs += ['NPs']
            param_sets += [self.nuisance_parameters]
        if 'global_observable' in items:
            param_strs += ['Global Observables']
            param_sets += [self.global_observables]
        for pstr, pset in zip(param_strs, param_sets):
            summary_str += f"{pstr} ({len(pset)}):\n"
            if include_patterns is not None:
                pset = [p for pattern in include_patterns for p in pset if fnmatch.fnmatch(p.GetName(), pattern)]
            if exclude_patterns is not None:
                pset = [p for pattern in exclude_patterns for p in pset if not fnmatch.fnmatch(p.GetName(), pattern)]
                
            data = [get_variable_attributes(p) for p in pset]
            data = sorted(data, key=lambda d: d['name'])
            if detailed:
                for d in data:
                    name   = d['name']
                    value  = d['value']
                    vmin   = d['min']
                    vmax   = d['max']
                    const_str = "C" if d['is_constant'] else "F"
                    summary_str += f"\t{name} = {value} [{vmin}, {vmax}] {const_str}\n"
            else:
                for d in data:
                    name = d['name']
                    summary_str += f"\t{name}\n"
        if not suppress_print:
            print(summary_str)
        if save_as is not None:
            with open(save_as, "w") as f:
                f.write(summary_str)


from typing import Optional, Union, Dict, List

import matplotlib.patches as patches
import matplotlib.lines as lines
import pandas as pd
import ROOT

from quickstats.plots.template import single_frame, parse_styles, format_axis_ticks, draw_atlas_label
from quickstats.plots import AbstractPlot
from quickstats.utils.common_utils import combine_dict

class UpperLimit2DPlot(AbstractPlot):
    
    COLOR_PALLETE = {
        '2sigma': '#FDC536',
        '1sigma': '#4AD9D9',
        'expected': 'k',
        'observed': 'k'
    }
    
    COLOR_PALLETE_SEC = {
        '2sigma': '#FDC536',
        '1sigma': '#4AD9D9',
        'expected': 'r',
        'observed': 'r'
    }
    
    LABELS = {
        '2sigma': 'Expected limit $\pm 2\sigma$',
        '1sigma': 'Expected limit $\pm 1\sigma$',
        'expected': 'Expected limit (95% CL)',
        'observed': 'Observed limit (95% CL)'
    }
    
    LABELS_SEC = {
        '2sigma': 'Expected limit $\pm 2\sigma$',
        '1sigma': 'Expected limit $\pm 1\sigma$',
        'expected': 'Expected limit (95% CL)',
        'observed': 'Observed limit (95% CL)'
    }

    CONFIG = {
        'primary_hatch'  : '\\\\\\',
        'secondary_hatch': '///',
        'primary_alpha'  : 0.9,
        'secondary_alpha': 0.8,
        'curve_line_styles': {
            'color': 'darkred' 
        },
        'curve_fill_styles':{
            'color': '#F2385A'
        },
        'highlight_styles': {
            'linewidth' : 0,
            'marker' : '*',
            'markersize' : 20,
            'color' : '#E9F1DF',
            'markeredgecolor' : 'black'
        }
    }
    
    def __init__(self, data, data_sec=None,
                 scale_factor=None,
                 color_pallete:Optional[Dict]=None,
                 color_pallete_sec:Optional[Dict]=None,
                 labels:Optional[Dict]=None,
                 labels_sec:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]='limit',
                 analysis_label_options:Optional[Union[Dict, str]]='default'):
        super().__init__(color_pallete=color_pallete,
                         color_pallete_sec=color_pallete_sec,
                         styles=styles,
                         analysis_label_options=analysis_label_options)
        self.data     = data
        # secondary data
        self.data_sec = data_sec
        
        self.labels = combine_dict(self.LABELS, labels)
        self.labels_sec = combine_dict(self.LABELS_SEC, labels_sec)
            
        self.scale_factor = scale_factor
        
        self.curve_data     = None
        self.highlight_data = None
    
        self.config = combine_dict(self.CONFIG)
        
    def get_default_legend_order(self):
        return ['observed', 'expected', '1sigma', '2sigma', 'curve', 'highlight']
    
    def add_curve(self, x, y, yerrlo=None, yerrhi=None,
                  label:str="Theory prediction",
                  line_styles:Optional[Dict]=None,
                  fill_styles:Optional[Dict]=None):
        curve_data = {
            'x'     : x,
            'y'     : y,
            'yerrlo'  : yerrlo,
            'yerrhi'  : yerrhi,
            'label' : label,
            'line_styles': line_styles,
            'fill_styles': fill_styles,
        }
        self.curve_data = curve_data
                        
    def add_highlight(self, x:float, y:float, label:str="SM prediction",
                      styles:Optional[Dict]=None):
        highlight_data = {
            'x'     : x,
            'y'     : y,
            'label' : label,
            'styles': styles
        }
        self.highlight_data = highlight_data        
    
    def draw_curve(self, ax, data):
        line_styles = data['line_styles']
        fill_styles = data['fill_styles']
        if line_styles is None:
            line_styles = self.config['curve_line_styles']
        if fill_styles is None:
            fill_styles = self.config['curve_fill_styles']
        handle_line = ax.plot(data['x'], data['y'], label=data['label'], **line_styles)
        handle_fill = ax.fill_between(data['x'], data['yerrlo'], data['yerrhi'],
                                      label=data['label'], **fill_styles)
        self.update_legend_handles({'curve': (handle_fill, handle_line[0])})
        
    def draw_highlight(self, ax, data):
        styles = data['styles']
        if styles is None:
            styles = self.config['highlight_styles']
        handle = ax.plot(data['x'], data['y'], label=data['label'], **styles)
        self.update_legend_handles({'highlight': handle[0]})
        
    def draw_single_data(self, ax, data, scale_factor=None,
                         log:bool=False, 
                         draw_observed:bool=True,
                         color_pallete:Optional[Dict]=None,
                         labels:Optional[Dict]=None,
                         observed_marker:Optional[str]='o', 
                         sigma_band_hatch:Optional[str]=None,
                         alpha:float=1.,
                         sec:bool=False):
        
        if color_pallete is None:
            color_pallete = self.color_pallete
        if labels is None:
            labels = self.labels
        if scale_factor is None:
            scale_factor = 1.0
            
        indices = data.index.astype(float).values
        exp_limits = data['0'].values * scale_factor
        n1sigma_limits = data['-1'].values * scale_factor
        n2sigma_limits = data['-2'].values * scale_factor
        p1sigma_limits = data['1'].values * scale_factor
        p2sigma_limits = data['2'].values * scale_factor
        
        handles_map = {}
        
        # draw +- 1, 2 sigma bands 
        handle_2sigma = ax.fill_between(indices, n2sigma_limits, p2sigma_limits, 
                                        facecolor=color_pallete['2sigma'],
                                        label=labels['2sigma'],
                                        hatch=sigma_band_hatch,
                                        alpha=alpha)
        handle_1sigma = ax.fill_between(indices, n1sigma_limits, p1sigma_limits, 
                                        facecolor=color_pallete['1sigma'],
                                        label=labels['1sigma'],
                                        hatch=sigma_band_hatch,
                                        alpha=alpha)
        handles_map['1sigma'] = handle_1sigma
        handles_map['2sigma'] = handle_2sigma
        
        if log:
            draw_fn = ax.semilogy
        else:
            draw_fn = ax.plot
 
        if draw_observed:
            obs_limits = data['obs'].values * scale_factor
            handle_observed = draw_fn(indices, obs_limits, color=color_pallete['observed'], 
                                      label=labels['observed'], 
                                      marker=observed_marker,
                                      alpha=alpha)
            handles_map['observed'] = handle_observed[0]
            
        handle_expected = draw_fn(indices, exp_limits, color=color_pallete['expected'],
                                  linestyle='--',
                                  label=labels['expected'],
                                  alpha=alpha)
        handles_map['expected'] = handle_expected[0]

        self.update_legend_handles(handles_map, sec=sec)
            
    def draw(self, xlabel:str="", ylabel:str="", ylim=None, xlim=None,
             log:bool=False, draw_observed:bool=True, observed_marker:Optional[str]='o'):
        ax = single_frame(styles=self.styles)

        if self.data_sec is not None:
            self.draw_single_data(ax, self.data_sec,
                                  scale_factor=self.scale_factor,
                                  log=log, 
                                  draw_observed=draw_observed,
                                  color_pallete=self.color_pallete_sec,
                                  labels=self.labels_sec,
                                  observed_marker=observed_marker, 
                                  sigma_band_hatch=self.config['secondary_hatch'],
                                  alpha=self.config['secondary_alpha'],
                                  sec=True)
            sigma_band_hatch = self.config['primary_hatch']
            alpha = self.config['primary_alpha']
        else:
            sigma_band_hatch = None
            alpha = 1.
        self.draw_single_data(ax, self.data,
                              scale_factor=self.scale_factor,
                              log=log, 
                              draw_observed=draw_observed,
                              color_pallete=self.color_pallete,
                              labels=self.labels,
                              observed_marker=observed_marker, 
                              sigma_band_hatch=sigma_band_hatch,
                              alpha=alpha)
        if self.curve_data is not None:
            self.draw_curve(ax, self.curve_data)   
        if self.highlight_data is not None:
            self.draw_highlight(ax, self.highlight_data)            
        ax.set_xlabel(xlabel, **self.styles['xlabel'])
        ax.set_ylabel(ylabel, **self.styles['ylabel'])
        format_axis_ticks(ax, **self.styles['axis'])
        
        if ylim is not None:
            ax.set_ylim(*ylim)
        if xlim is not None:
            ax.set_xlim(*xlim)
        
        if self.analysis_label_options is not None:
            draw_atlas_label(ax, dy=0.05, text_options=self.styles['text'], **self.analysis_label_options)
        # border for the legend
        border_leg = patches.Rectangle((0, 0), 1, 1, facecolor = 'none', edgecolor = 'black', linewidth = 1)
        self.legend_data['1sigma']['handle'] = (self.legend_data['1sigma']['handle'], border_leg)
        self.legend_data['2sigma']['handle'] = (self.legend_data['2sigma']['handle'], border_leg)
        if self.curve_data is not None:
            self.legend_data['curve']['handle'] = (*self.legend_data['curve']['handle'], border_leg)
        
        if self.data_sec is not None:
            self.legend_data_sec['1sigma']['handle'] = (self.legend_data_sec['1sigma']['handle'], border_leg)
            self.legend_data_sec['2sigma']['handle'] = (self.legend_data_sec['2sigma']['handle'], border_leg)
            
        handles, labels = self.get_legend_handles_labels()
        if self.data_sec is not None:
            handles_sec, labels_sec = self.get_legend_handles_labels(sec=True)
            handles = handles + handles_sec
            labels  = labels + labels_sec
        ax.legend(handles, labels, **self.styles['legend'])
        return ax


from typing import Optional, Dict, List, Union
import enum

from quickstats.components import AnalysisObject
from quickstats.utils.common_utils import parse_config

class AsimovType(enum.Enum):
    S_NP_Nom               = -2
    B_NP_Nom               = -1
    B_NP_Fit               = 0
    S_NP_Fit               = 1
    S_NP_Fit_muhat         = 2
    B_unconstrained_NP_Fit = 3
    S_unconstrained_NP_Fit = 4
    S_unconstrained_NP_Fit_muhat = 5
    
class AsimovGenerator(AnalysisObject):

    ASIMOV_SETTINGS = {
        AsimovType.S_NP_Nom: {
            "asimov_name": "asimovData_1_NP_Nominal",
            "asimov_snapshot": "asimovData_1_NP_Nominal",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": False,
            "modify_globs": False
        },
        AsimovType.B_NP_Nom: {
            "asimov_name": "asimovData_0_NP_Nominal",
            "asimov_snapshot": "asimovData_0_NP_Nominal",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": False,
            "modify_globs": False
        },
        AsimovType.B_NP_Fit: {
            "asimov_name": "asimovData_0_NP_Profile",
            "asimov_snapshot": "asimovData_0_NP_Profile",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.S_NP_Fit: {
            "asimov_name": "asimovData_1_NP_Profile",
            "asimov_snapshot": "asimovData_1_NP_Profile",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.S_NP_Fit_muhat: {
            "asimov_name": "asimovData_muhat_NP_Profile",
            "asimov_snapshot": "asimovData_muhat_NP_Profile",
            "poi_val": 1,
            "poi_profile": None,
            "do_fit": True,
            "modify_globs": True
        },
        AsimovType.B_unconstrained_NP_Fit: {
            "asimov_name": "asimovData_0_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_0_unconstrained_NP_Profile",
            "poi_val": 0,
            "poi_profile": 0,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        },
        AsimovType.S_unconstrained_NP_Fit: {
            "asimov_name": "asimovData_1_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_1_unconstrained_NP_Profile",
            "poi_val": 1,
            "poi_profile": 1,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        },
        AsimovType.S_unconstrained_NP_Fit_muhat: {
            "asimov_name": "asimovData_muhat_unconstrained_NP_Profile",
            "asimov_snapshot": "asimovData_muhat_unconstrained_NP_Profile",
            "poi_val": 1,
            "poi_profile": None,
            "do_fit": True,
            "modify_globs": True,
            "constraint_option": 1
        }
    }    
    
    DEFAULT_ASIMOV_TYPES = [AsimovType.B_NP_Fit, AsimovType.S_NP_Fit, AsimovType.S_NP_Fit_muhat]
    
    def __init__(self, filename:str, poi_name:str=None,
                 data_name:str='combData', 
                 config:Optional[Union[Dict, str]]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        config = parse_config(config)
        config['filename']  = filename
        config['poi_name']  = poi_name
        config['data_name'] = data_name
        config['verbosity'] = verbosity
        self._inherit_init(super(AsimovGenerator, self).__init__, **config)
        
    
    def generate_standard_asimov(self, asimov_types:Optional[Union[List[AsimovType], List[int]]]=None,
                                 asimov_names:Optional[Union[List[str], str]]=None,
                                 asimov_snapshots:Optional[Union[List[str], str]]=None,
                                 poi_scale:Optional[float]=None):
        if asimov_types is None:
            asimov_types = self.DEFAULT_ASIMOV_TYPES
        if isinstance(asimov_types, (int, AsimovType)):
            asimov_types = [asimov_types]
        if asimov_names is not None:
            if isinstance(asimov_names, str):
                asimov_names = [asimov_names]
            assert len(asimov_names) == len(asimov_types)
        if asimov_snapshots is not None:
            if isinstance(asimov_snapshots, str):
                asimov_snapshots = [asimov_snapshots]
            assert len(asimov_snapshots) == len(asimov_types)            
        if poi_scale is None:
            poi_scale = 1.0
        poi_name = self.poi.GetName()
        self.model.workspace.saveSnapshot("temp", self.model.workspace.allVars())
        for i, asimov_type in enumerate(asimov_types):
            self.load_snapshot("temp")
            if isinstance(asimov_type, int):
                asimov_type = AsimovType(asimov_type)
            kwargs = self.ASIMOV_SETTINGS[asimov_type]
            kwargs['poi_name'] = poi_name
            for key in ['poi_val', 'poi_profile']:
                if kwargs[key] is not None:
                    kwargs[key] *= poi_scale
            if asimov_names is not None:
                kwargs['asimov_name'] = asimov_names[i]
            if asimov_snapshots is not None:
                kwargs['asimov_snapshot'] = asimov_snapshots[i]
            kwargs['minimizer_options'] = self.minimizer_options
            kwargs['nll_options'] = self.nll_commands
            self.model.generate_asimov(**kwargs)


from typing import Optional, Union, Dict, List
import os
import re
import copy
import glob
import itertools

import numpy as np

from quickstats.components.numerics import pretty_float, str_encode_value, str_decode_value


signature_regex = {
    'F': r"\d+[.]?\d*",
    'P': r"n?\d+p?\d*",
    'S': r"\w+"
}

signature_parser = {
    'F': pretty_float,
    'P': str_decode_value,
    'S': str
}

signature_regex = {
    'F': r"\d+[.]?\d*",
    'P': r"n?\d+p?\d*",
    'S': r"\w+"
}

signature_parser = {
    'F': pretty_float,
    'P': str_decode_value,
    'S': str
}

class ParamParser:
    
    DEFAULT_FORMAT_STR = r"\w+"
    
    @property
    def format_str(self):
        return self._format_str
    @format_str.setter
    def format_str(self, val):
        if val is None:
            self._format_str = self.DEFAULT_FORMAT_STR
        else:
            self._format_str = val
        self.fname_regex   = self.get_fname_regex(self._format_str)
        self.attribute_parser = self.get_attribute_parser(self._format_str)

    def __init__(self, format_str:str=None, internal_param_str:str=None):
        self.format_str   = format_str
        self.internal_param_str = internal_param_str
    
    @staticmethod
    def get_signature_map(format_str:str):
        attribute_groups = re.findall(r"<(\w+)\[(\w)\]>", format_str)
        signature_map = {}
        for group in attribute_groups:
            attribute, signature = group[0], group[1]
            signature_map[attribute] = signature
        return signature_map
    
    @staticmethod
    def get_fname_regex(format_str:str, ext:str=".root"):
        expr = format_str
        signature_map = ParamParser.get_signature_map(format_str)
        for attribute, signature in signature_map.items():
            attribute_expr = signature_regex.get(signature.upper(), None)
            if expr is None:
                raise ValueError(f"unknown signature `{signature}`")
            group_expr = f"(?P<{attribute}>{attribute_expr})"
            expr = expr.replace(f"<{attribute}[{signature}]>", group_expr)
        expr += (ext.replace('.', r'\.') + "$")
        regex = re.compile(expr)
        return regex
   
    @staticmethod
    def sort_param_points(param_points:List, attributes:List):
        key = lambda d: tuple(d['parameters'][attrib] for attrib in attributes)
        return sorted(param_points, key=key)
    
    @staticmethod
    def get_attribute_parser(format_str:str):
        attribute_parser = {}
        signature_map = ParamParser.get_signature_map(format_str)
        for attribute, signature in signature_map.items():
            parser = signature_parser.get(signature, None)
            if parser is None:
                raise ValueError(f"unknown signature `{signature}`")
            attribute_parser[attribute] = parser
        return attribute_parser

    @staticmethod
    def parse_param_str(param_str:Optional[str]=None):
        if param_str is None:
            return {}
        param_values = {}
        param_expr_list = [s.strip() for s in param_str.split(',')]
        param_expr_list = [s for s in param_expr_list if s]
        for expr in param_expr_list:
            tokens = expr.split('=')
            if len(tokens) != 2:
                raise ValueError('invalid expression for parameterisation')
            param_name = tokens[0]
            values_expr = tokens[1]
            tokens = values_expr.split('_')
            # fixed value
            if len(tokens) == 1:
                values = [float(tokens[0])]
            # scan across range
            elif len(tokens) == 3:
                poi_min = float(tokens[0])
                poi_max = float(tokens[1])
                poi_step = float(tokens[2])
                values = np.arange(poi_min, poi_max + poi_step, poi_step)
            else:
                raise ValueError('invalid expression for parameterisation')
            param_values[param_name] = values
        param_names = list(param_values.keys())
        combinations = [param_values[param_name] for param_name in param_names]
        combinations = itertools.product(*combinations)
        param_points = []
        for combination in combinations:
            param_point = {k:v for k,v in zip(param_names, combination)}
            param_points.append(param_point)
        return param_points
    
    @staticmethod
    def val_encode_parameters(parameters:Dict):
        return ",".join([f"{param}={round(value, 8)}" for param, value in parameters.items()])
    
    @staticmethod
    def str_encode_parameters(parameters:Dict):
        encoded_str_list = []
        for param, value in parameters.items():
            if isinstance(value, float):
                value = str_encode_value(round(value, 8))
            encoded_str = f"{param}_{value}"
            encoded_str_list.append(encoded_str)
        return "_".join(encoded_str_list)

    def get_external_param_points(self, dirname:str=""):
        fnames = glob.glob(os.path.join(dirname, '*'))
        param_points = []
        for fname in fnames:
            basename = os.path.basename(fname)
            match = self.fname_regex.match(basename)
            if not match:
                continue
            point = {'filename': fname, 'parameters':{}}
            point['basename'] = basename.split('.')[0]
            for key, value in match.groupdict().items():
                parser = self.attribute_parser[key]
                point['parameters'][key] = parser(value)
            param_points.append(point)
        attributes = list(self.attribute_parser)
        param_points = self.sort_param_points(param_points, attributes)
        return param_points
    
    def get_internal_param_points(self):
        return self.parse_param_str(self.internal_param_str)
    
    def get_param_points(self, dirname:str=""):
        external_param_points = self.get_external_param_points(dirname)     
        internal_param_points = self.get_internal_param_points()
        if len(internal_param_points) == 0:
            internal_param_points = [{}]
        param_points = []
        for ext_point in external_param_points:
            fname = ext_point['filename']
            basename = ext_point['basename']
            ext_params = ext_point['parameters']
            for int_params in internal_param_points:
                param_point = {}
                param_point['filename'] = fname
                param_point['basename'] = basename
                param_point['external_parameters'] = {**ext_params}
                param_point['internal_parameters'] = {**int_params}
                param_points.append(param_point)
        return param_points


from typing import List, Union, Optional, Dict
import os
import glob

import quickstats

def compile_macros(macros:Optional[Union[str, List[str]]]=None):
    """
    Compile ROOT macros
    
    Arguments:
        macros: (Optional) str or list of str
            If str, it is a string containing comma delimited list of macro names to be compiled.
            If list of str, it is the list of macro names to be compiled.
    """
    if macros is None:
        macros = get_all_macros()
    elif isinstance(macros, str):
        macros = macros.split(',')
    from quickstats.utils.root_utils import compile_macro
    for macro in macros:
        compile_macro(macro)
        
def get_all_macros():
    """
    Get the list of macros names
    
    Note: Only macros ending in .cxx will be considered
    """
    macros_dir = os.path.join(quickstats.macro_path, 'macros')
    macro_paths = glob.glob(os.path.join(macros_dir, "*.cxx"))
    macros = [os.path.os.path.splitext(os.path.basename(path))[0] for path in macro_paths]
    return macros


def set_verbosity(verbosity:Union[str, int]="INFO"):
    quickstats._PRINT_.verbosity = verbosity


from typing import Optional, List, Dict
import fnmatch
import os
import re
import ROOT

import numpy as np
import pandas as pd

class RooProcBaseAction(object):
    def __init__(self, **params):
        self._params = params
        self.executed = False
        self.status   = None
    
    def get_formatted_parameters(self, global_vars:Optional[Dict]=None):
        if global_vars is None:
            global_vars = {}
        formatted_parameters = {}
        for k,v in self._params.items():
            if v is None:
                formatted_parameters[k] = None
                continue
            k_literals = re.findall(r"\${(\w+)}", k)
            if isinstance(v, list):
                v = '__SEPARATOR__'.join(v)
            v_literals = re.findall(r"\${(\w+)}", v)
            all_literals = set(k_literals).union(set(v_literals))
            for literal in all_literals:
                if literal not in global_vars:
                    raise RuntimeError(f"the substitute literal `{literal}` is undefined")
            for literal in k_literals:
                substitute = global_vars[literal]
                k = k.replace("${" + literal + "}", str(substitute))
            for literal in v_literals:
                substitute = global_vars[literal]
                v = v.replace("${" + literal + "}", str(substitute))
            if '__SEPARATOR__' in v:
                v = v.split("__SEPARATOR__")
            formatted_parameters[k] = v
        return formatted_parameters
    
    def makedirs(self, fname):
        dirname = os.path.dirname(fname)
        if dirname and (not os.path.exists(dirname)):
            os.makedirs(dirname)
    
    def execute(self, **params):
        raise NotImplementedError
    
    @classmethod
    def parse_as_kwargs(cls, text):
        kwargs = {}
        text = re.sub(r"\s*", "", text)
        list_attributes = re.findall(r"(\w+)=\[([^\[\]]+)\]", text)
        for attribute in list_attributes:
            kwargs[attribute[0]] = attribute[1].split(",")
            text = text.replace(f"{attribute[0]}=[{attribute[1]}]","")
        attributes = re.findall(r"(\w+)=([^,]+)", text)
        for attribute in attributes:
            kwargs[attribute[0]] = attribute[1]
        return kwargs
    
    @classmethod
    def parse(cls, text):
        raise NotImplementedError
        
class RooProcRDFAction(RooProcBaseAction):
    def _execute(self, rdf, **params):
        return rdf
    
    def execute(self, rdf, global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        rdf_next = self._execute(rdf, **params)
        self.executed = True
        return rdf_next         
        
class RooProcHelperAction(RooProcBaseAction):
    def _execute(self, processor, **params):
        return processor
    
    def execute(self, processor, global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        processor = self._execute(processor, **params)
        self.executed = True
        return processor

class RooProcHybridAction(RooProcBaseAction):
    
    def _execute(self, rdf, processor, **params):
        return rdf, processor
    
    def execute(self, rdf, processor, global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        rdf_next, processor_next = self._execute(rdf, processor, **params)
        self.executed = True
        return rdf_next, processor_next
    
    
class RooProcTreeName(RooProcHelperAction):
    
    def __init__(self, treename:str):
        super().__init__(treename=treename)
        
    @classmethod
    def parse(cls, text):
        return cls(treename=text)
    
    def _execute(self, processor, **params):
        treename = params['treename']
        processor.treename = treename
        return processor
    
class RooProcExport(RooProcHelperAction):
    def __init__(self, filename:str):
        super().__init__(filename=filename)
        
    @classmethod
    def parse(cls, text):
        kwargs = cls.parse_as_kwargs(text)
        return cls(**kwargs)
    
    def _execute(self, processor, **params):
        filename = kwargs['filename']
        data = {k:v.GetValue() for k,v in processor.external_variables.items()}
        dirname = os.path.dirname(filename)
        if dirname and (not os.path.exists(dirname)):
            os.makedirs(dirname)
        with open(filename, 'w') as outfile:
            json.dump(data, outfile, indent=2)
        return processor
    
class RooProcGlobalVariables(RooProcHelperAction):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    @classmethod
    def parse(cls, text):
        text = re.sub(r"\s*", "", text)
        tokens = text.split(",")
        globs = {}
        for token in tokens:
            result = re.match("^(\w+)=(.*)", token)
            if not result:
                raise RuntimeError(f"invalid expression {token}")
            globs[result[1]] = result[2]
        return cls(**globs)
    
    def _execute(self, processor, **params):
        processor.global_variables.update(params)
        return processor
    
class RooProcDefine(RooProcRDFAction):
    
    def __init__(self, name:str, expression:str):
        super().__init__(name=name, expression=expression)
        
    @classmethod
    def parse(cls, text):
        result = re.search(r"^\s*(\w+)\s*=(.*)", text)
        if not result:
            raise RuntimeError(f"invalid expression {text}")
        name = result.group(1)
        expression = result.group(2)
        return cls(name=name, expression=expression)        
        
    def _execute(self, rdf, **params):
        name = params['name']
        expression = params['expression']
        rdf_next = rdf.Define(name, expression)
        return rdf_next
    
class RooProcFilter(RooProcRDFAction):
    def __init__(self, expression:str, name:Optional[str]=None):
        super().__init__(expression=expression,
                         name=name)
        
    @classmethod
    def parse(cls, text):
        name_literals = re.findall(r"@{([^{}]+)}", text)
        if len(name_literals) == 0:
            name = None
            expression = text.strip()
        elif len(name_literals) == 1:
            name = name_literals[0]
            expression = text.replace("@{" + name + "}", "").strip()
        else:
            raise RuntimeError(f"multiple filter names detected in the expression `{text}`")
        return cls(name=name, expression=expression)
        
    def _execute(self, rdf, **params):
        expression = params['expression']
        name = params.get("name", None)
        if name is not None:
            rdf_next = rdf.Filter(expression, name)
        else:
            rdf_next = rdf.Filter(expression)
        return rdf_next
    
class RooProcSave(RooProcHybridAction):
    
    def __init__(self, treename:str, filename:str, 
                 columns:Optional[List[str]]=None,
                 frame:Optional[str]=None):
        super().__init__(treename=treename,
                         filename=filename,
                         columns=columns,
                         frame=frame)
        
    @classmethod
    def parse(cls, text):
        kwargs = cls.parse_as_kwargs(text)
        return cls(**kwargs)
    
    def _execute(self, rdf, processor, **params):
        treename = params['treename']
        filename = params['filename']
        columns = params.get('columns', None)
        if columns is None:
            rdf_next = rdf.Snapshot(treename, filename)
        else:
            all_columns = [str(c) for c in rdf.GetColumnNames()]
            save_columns = []
            for column in columns:
                save_columns += [c for c in all_columns if fnmatch.fnmatch(c, column)]
            save_columns = list(set(save_columns))
            self.makedirs(filename)
            rdf_next = rdf.Snapshot(treename, filename, save_columns)
        processor.stdout.info(f"INFO: Writing output to `{filename}`.")
        return rdf_next, processor
    
class RooProcAsNumpy(RooProcRDFAction):
    
    def __init__(self, filename:str, 
                 columns:Optional[List[str]]=None):
        super().__init__(filename=filename,
                         columns=columns)
        
    @classmethod
    def parse(cls, text):
        kwargs = cls.parse_as_kwargs(text)
        return cls(**kwargs)
    
    def _execute(self, rdf, **kwargs):
        filename = kwargs['filename']
        columns = kwargs.get('columns', None)
        if columns is None:
            data = rdf.AsNumpy()
        else:
            data = rdf.AsNumpy(columns)
        np.save(filename, data)
        return rdf    
    
class RooProcReport(RooProcHybridAction):
    def __init__(self, display:bool=False, filename:Optional[str]=None):
        super().__init__(display=display,
                         filename=filename)
    @classmethod
    def parse(cls, text):
        kwargs = cls.parse_as_kwargs(text)
        return cls(**kwargs)
    
    def _execute(self, rdf, processor, **params):
        display = params['display']
        filename = params['filename']
        cut_report = rdf.Report()
        result = []
        cumulative_eff  = 1
        for report in cut_report:
            data = {}
            data['name'] = report.GetName()
            data['all']  = report.GetAll()
            data['pass'] = report.GetPass()
            data['efficiency'] = report.GetEff()
            cumulative_eff *= data['efficiency']/100
            data['cumulative_efficiency'] = cumulative_eff*100
            result.append(data)
        df = pd.DataFrame(result)
        if int(display):
            processor.stdout.info(df)
        if filename is not None:
            self.makedirs(filename)
            df.to_csv(filename)
        return rdf, processor
    
class RooProcSum(RooProcHybridAction):
    def __init__(self, ext_var_name:str, column_name:str):
        super().__init__(ext_var_name=ext_var_name,
                         column_name=column_name)
        
    @classmethod
    def parse(cls, text):
        name_literals = re.findall(r"@{([^{}]+)}", text)
        if len(name_literals) == 1:
            ext_var_name = name_literals[0]
            column_name = text.replace("@{" + ext_var_name + "}", "").strip()
        else:
            raise RuntimeError("unspecified external variable name (format:@{ext_var_name})")
        return cls(ext_var_name=ext_var_name, column_name=column_name)
        
    def _execute(self, rdf, processor, **params):
        ext_var_name = params['ext_var_name']
        column_name = params['column_name']
        processor.external_variables[ext_var_name] = rdf.Sum(column_name)
        return rdf, processor
        
class RooProcDeclare(RooProcHelperAction):
    
    def __init__(self, expression:str, name:Optional[str]=None):
        super().__init__(expression=expression,
                         name=name)
        
    @classmethod
    def parse(cls, text):
        return cls(expression=text)        
        
    @staticmethod
    def create_declaration(expression:str, name:Optional[str]=None):
        if name is None:
            hash_str = str(hash(expression)).replace("-", "n")
            name_guard = f"__RooProcDeclare_{hash_str}__"
        else:
            name_guard = f"__RooProcDeclare_{name}__"
        guarded_declaration = f"#ifndef {name_guard}\n"
        guarded_declaration += f"#define {name_guard}\n"
        guarded_declaration += f"\n{expression}\n\n#endif\n"
        return guarded_declaration
    
    @staticmethod
    def declare_expression(expression:str, name:Optional[str]=None):
        declaration = RooProcDeclare.create_declaration(expression, name)
        status = ROOT.gInterpreter.Declare(declaration)
        return status
    
    def _execute(self, processor, **params):
        name = params.get("name", None)
        expression = params['expression']
        self.declare_expression(expression, name)
        return processor
    
class RooProcSaveFrame(RooProcHelperAction):
    def __init__(self, name:str):
        super().__init__(name=name)

    @classmethod
    def parse(cls, text):
        return cls(name=text)
    
    def _execute(self, processor, **params):
        frame_name = params['name']
        if frame_name in processor.rdf_frames:
            processor.stdout.warning(f"WARNING: Overriding existing rdf frame `{frame_name}`")
        processor.rdf_frames[frame_name] = processor.rdf
        return processor
    
class RooProcLoadFrame(RooProcHelperAction):
    def __init__(self, name:str):
        super().__init__(name=name)

    @classmethod
    def parse(cls, text):
        return cls(name=text)
    
    def _execute(self, processor, **params):
        frame_name = params['name']
        if frame_name not in processor.rdf_frames:
            raise RuntimeError(f"failed to load rdf frame `{frame_name}`: frame does not exist.")
        processor.rdf = processor.rdf_frames[frame_name]

ACTION_MAP = {
    'TREENAME': RooProcTreeName,
    'DECLARE': RooProcDeclare,
    'GLOBAL': RooProcGlobalVariables,
    'DEFINE': RooProcDefine,
    'FILTER': RooProcFilter,
    'SAVE': RooProcSave,
    'REPORT': RooProcReport,
    'GETSUM': RooProcSum,
    'EXPORT': RooProcExport,
    'SAVE_FRAME': RooProcSaveFrame,
    'LOAD_FRAME': RooProcLoadFrame,
    'AS_NUMPY': RooProcAsNumpy
}


import re

from .actions import ACTION_MAP

class RooProcConfigParser(object):
    
    kActions      = tuple(ACTION_MAP)
    
    def __init__(self):
        self.action_list = []
    
    @staticmethod
    def get_start_tag(text:str):
        result = re.search(r"^\s*<(\w+)>", text)
        if not result:
            return None
        else:
            return "<"+result.group(1)+">"
        
    @staticmethod
    def get_end_tag(text:str):
        result = re.search(r"</(\w+)>\s*$", text)
        if not result:
            return None
        else:
            return "</"+result.group(1)+">"
        
    @staticmethod
    def parse_text(text:str):
        lines = text.splitlines(True)
        # remove comments
        lines = [l.split("#")[0] for l in lines]
        # remove blank lines
        lines = filter(lambda x: not re.match(r'^\s*$', x), lines)
        action_list = []
        action_name = None
        expression  = ""
        block_flag  = False
        for l in lines:
            if action_name is None:
                tokens = [t.strip() for t in l.strip().split()]
                # remove empty tokens
                tokens = [t for t in tokens if t]
                if not tokens:
                    continue
                action_name = tokens[0].upper()
                if action_name not in RooProcConfigParser.kActions:
                    raise RuntimeError(f"unknown action `{action_name}`")
                if (len(tokens) == 1):
                    continue
                else:
                    expression = ' '.join(tokens[1:])
            else:
                start_tag = RooProcConfigParser.get_start_tag(l)
                end_tag = RooProcConfigParser.get_end_tag(l)
                if start_tag:
                    if block_flag:
                        raise RuntimeError("missing end-tag or multiple start-tags detected")
                    block_flag = True
                    l = l.split(start_tag)[1]
                if end_tag:
                    if not block_flag:
                        raise RuntimeError("missing start-tag or multiple end-tags detected")
                    block_flag = False
                    l = l.split(end_tag)[0]
                expression += l
            if block_flag:
                continue
            else:
                expression = expression.strip()
                action = RooProcConfigParser.create_action(action_name, expression)
                action_list.append(action)
                action_name = None
                expression  = ""
        if block_flag:
            raise RuntimeError("unterminated start-tag detected")
        return action_list
    
    @staticmethod
    def parse_file(path:str):
        with open(path, "r") as f:
            text = f.read()
        return RooProcConfigParser.parse_text(text)
    
    @staticmethod
    def create_action(action_name:str, expression:str):
        cls = ACTION_MAP.get(action_name, None)
        if cls is None:
            raise RuntimeError(f"unknown action `{action_name}`")
        action = cls.parse(expression)
        return action


import ROOT

def set_attrib(obj, **kwargs):
    for key, value in kwargs.items():
        target = obj
        if '.' in key:
            tokens = key.split('.')
            if len(tokens) != 2:
                raise ValueError('maximum of 1 subfield is allowed but {} is given'.format(len(tokens)-1))
            field, key = tokens[0], tokens[1]
            method_name = 'Get' + field
            if hasattr(obj, 'Get' + field):
                target = getattr(obj, method_name)()
            else:
                raise ValueError('{} object does not contain the method {}'.format(type(target), method_name)) 
        method_name = 'Set' + key
        if hasattr(target, 'Set' + key):
            method_name = 'Set' + key
        elif hasattr(target, key):
            method_name = key
        else:
            raise ValueError('{} object does not contain the method {}'.format(type(target), method_name))         
        if value is None:
            getattr(target, method_name)()
        elif isinstance(value, (list, tuple)):
            getattr(target, method_name)(*value)
        else:
            getattr(target, method_name)(value)
    return obj


import functools
from concurrent.futures  import ProcessPoolExecutor

import quickstats

def parallel_run(func, *iterables, max_workers=None):
    max_workers = max_workers or quickstats.MAX_WORKERS 

    with ProcessPoolExecutor(max_workers) as executor:
        result = executor.map(func, *iterables)

    return [i for i in result]


from quickstats._version import __version__
from quickstats.decorators import *

import os
import pathlib

import ROOT
ROOT.gROOT.SetBatch(True) 
ROOT.PyConfig.IgnoreCommandLineOptions = True
module_path = pathlib.Path(__file__).parent.absolute()
macro_path = module_path
stylesheet_path = os.path.join(module_path, 'stylesheets')

# ROOT.gInterpreter.AddIncludePath(os.path.join(macro_path, "macros"))

MAX_WORKERS = 8

from quickstats.utils.io import VerbosePrint

_PRINT_ = VerbosePrint("INFO")

from quickstats.main import *

from quickstats.components.asimov_generator import AsimovType





class LikelihoodScanPlot(object):
    def __init__(self):
        self._data = None
        
    @property
    def data(self):
        return self._data
    
    def load(self, input_files):
        pass
             


##################################################################################################
# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools by Stefan Gadatsch
# Author: Alkaid Cheng
# Email: chi.lung.cheng@cern.ch
##################################################################################################
import os
import sys
import time
import json
import fnmatch
from itertools import repeat
from typing import List, Optional

import uproot
import numpy as np

from quickstats.components import ExtendedModel, ExtendedMinimizer
from quickstats.utils import root_utils, common_utils

import ROOT

class NuisanceParameterPull(object):
    @property
    def model(self):
        return self._model
    
    @property
    def workspace(self):
        return self._workspace
    @property
    def model_config(self):
        return self._model_config
    @property
    def pdf(self):
        return self._pdf
    @property
    def data(self):
        return self._data
    @property
    def nuisance_parameters(self):
        return self._nuisance_parameters
    @property
    def global_observables(self):
        return self._global_observables
    @property
    def pois(self):
        return self._pois
    @property
    def observables(self):
        return self._observables  
    
    def __init__(self):
        self._model               = None
        self._workspace           = None
        self._model_config        = None
        self._pdf                 = None
        self._data                = None
        self._nuisance_parameters = None
        self._global_observables  = None
        self._pois                = None
        self._observables         = None
    
    @staticmethod
    def evaluate_impact(model:ExtendedModel, minimizer:ExtendedMinimizer,
                        nuis, nominal_value, pois, minimizer_options, snapshot=None):
        poi_values = []
        start_time = time.time()
        if snapshot:
            model.workspace.loadSnapshot(snapshot)
        nuis.setVal(nominal_value)
        nuis.setConstant(1)   
        minimizer.minimize(**minimizer_options)
        for poi in pois:
            poi_values.append(poi.getVal())
        elapsed_time = time.time() - start_time
        return poi_values, elapsed_time

    @staticmethod
    def _run_pulls(input_file:str, data:str, snapshot:str, nuis_name:str, poi_name:str, 
                   fix_param:str='', profile_param:str='', binned:bool=True,
                   workspace:Optional[str]=None, model_config:Optional[str]=None,
                   minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                   precision:float=0.001, eps:float=1.0, strategy:int=0, print_level:int=1,
                   num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False,
                   fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                   max_iters:int=-1, batch_mode:bool=False, int_bin_precision:float=-1.,
                   outdir:str='pulls', cache=True, logfile_path=None, **kwargs):
        # create output directory if not exists
        if not os.path.exists(outdir):
            os.makedirs(outdir, exist_ok=True)        
        outname_root = os.path.join(outdir, nuis_name + '.root')
        outname_json = os.path.join(outdir, nuis_name + '.json')
        #checkpointing:
        if (os.path.exists(outname_root) and os.path.exists(outname_json)) and cache:
            print('INFO: Jobs already finished for the NP: {}. Skipping.'.format(nuis_name))
            return None
        else:
            print('INFO: Evaluating pulls for the NP: {}'.format(nuis_name))
        if logfile_path is not None:
            common_utils.redirect_stdout(logfile_path)
        start_time = time.time()
        # load model
        model = ExtendedModel(fname=input_file, ws_name=workspace,
                              mc_name=model_config, data_name=data, 
                              snapshot_name=snapshot, binned_likelihood=binned,
                              tag_as_measurement="pdf_", fix_cache=fix_cache, fix_multi=fix_multi)
        if fix_param:
            model.fix_parameters(fix_param)
            
        # by default fix all POIs before floating
        model.set_parameter_defaults(model.pois, error=0.15, constant=1, remove_range=True)
        for param in model.pois:
            extra_str = 'Fixing' if param.isConstant() else 'Set'
            print('INFO: {} POI {} at value {}'.format(extra_str, param.GetName(), param.getVal()))
        
        # collect pois
        rank_pois = model.profile_parameters(poi_name)
        model.set_parameter_defaults(rank_pois, error=0.15)     
        
        # profile pois
        if profile_param:
            print('INFO: Profiling POIs')
            profile_pois = model.profile_parameters(profile_param)
        
        buffer_time = time.time()
    
        nuip = model.workspace.var(nuis_name)
        if not nuip:
            raise ValueError('Nuisance parameter "{}" does not exist'.format(parameter))
        nuip_name = nuip.GetName()
        nuip.setConstant(0)
        print('INFO: Computing error for parameter "{}" at {}'.format(nuip.GetName(), nuip.getVal()))
        
        print("INFO: Making ExtendedMinimizer for unconditional fit")
        minimizer = ExtendedMinimizer("minimizer", model.pdf, model.data)
        print("INFO: Starting minimization")
        nll_commands = [ROOT.RooFit.NumCPU(num_cpu, 3),
                        #ROOT.RooFit.Constrain(model.nuisance_parameters),
                        ROOT.RooFit.GlobalObservables(model.global_observables), 
                        ROOT.RooFit.Offset(offset)]
        if hasattr(ROOT.RooFit, "BatchMode"):
            nll_commands.append(ROOT.RooFit.BatchMode(batch_mode))
        if hasattr(ROOT.RooFit, "IntegrateBins"):
            nll_commands.append(ROOT.RooFit.IntegrateBins(int_bin_precision))            

        minimize_options = {
            'minimizer_type'   : minimizer_type,
            'minimizer_algo'   : minimizer_algo,
            'default_strategy' : strategy,
            'opt_const'        : optimize,
            'precision'        : precision,
            'eps'              : eps,
            'max_calls'        : max_calls,
            'max_iters'        : max_iters,
        }

        minimizer.minimize(nll_commands=nll_commands,
                           scan=1, scan_set=ROOT.RooArgSet(nuip),
                           **minimize_options)
        unconditional_time = time.time() - buffer_time
        print("INFO: Fitting time: {:.3f} s".format(unconditional_time))
        pois_hat = []
        for rank_poi in rank_pois:
            name = rank_poi.GetName()
            value = rank_poi.getVal()
            pois_hat.append(value)
            print('{} {}'.format(name, value))
        
        model.workspace.saveSnapshot('tmp_snapshot', model.pdf.getParameters(model.data))
        print('INFO: Made unconditional snapshot with name tmp_snapshot')
        
        # find prefit variation
        buffer_time = time.time()
        
        nuip_hat = nuip.getVal()
        nuip_errup = nuip.getErrorHi()
        nuip_errdown = nuip.getErrorLo()

        all_constraints = model.get_all_constraints()
        prefit_variation, constraint_type, nuip_nom = model.inspect_constrained_nuisance_parameter(nuip, all_constraints)
        if not constraint_type:
            print('INFO: Not a constrained parameter. No prefit impact can be determined. Use postfit impact instead.')
        prefit_uncertainty_time = time.time() - buffer_time
        print('INFO: Time to find prefit variation: {:.3f} s'.format(prefit_uncertainty_time))
        
        if rank_pois:
            new_minimizer_options = {
                'nll_commands': nll_commands,
                'reuse_nll'   : 1,
                **minimize_options
            }
            # fix theta at the MLE value +/- postfit uncertainty and minimize again to estimate the change in the POI
            print('INFO: Evaluating effect of moving {} up by {} + {}'.format(nuip_name, nuip_hat, nuip_errup))
            pois_up, postfit_up_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                nuip, nuip_hat + abs(nuip_errup), rank_pois,
                                                new_minimizer_options,  'tmp_snapshot')
            print('INFO: Time to find postfit up impact: {:.3f} s'.format(postfit_up_impact_time))
            
            print('INFO: Evaluating effect of moving {} down by {} - {}'.format(nuip_name, nuip_hat, nuip_errup))
            pois_down, postfit_down_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                    nuip, nuip_hat - abs(nuip_errdown), rank_pois,
                                                    new_minimizer_options,  'tmp_snapshot')
            print('INFO: Time to find postfit down impact: {:.3f} s'.format(postfit_down_impact_time))
            
            # fix theta at the MLE value +/- prefit uncertainty and minimize again to estimate the change in the POI
            
            if constraint_type:
                print('Evaluating effect of moving {} up by {} + {}'.format(nuip_name, nuip_hat, prefit_variation))
                pois_nom_up, prefit_up_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                        nuip, nuip_hat + prefit_variation, rank_pois,
                                                        new_minimizer_options,  'tmp_snapshot')
                print('INFO: Time to find prefit up impact: {:.3f} s'.format(prefit_up_impact_time))      
                
                print('Evaluating effect of moving {} down by {} - {}'.format(nuip_name, nuip_hat, prefit_variation))
                pois_nom_down, prefit_down_impact_time = NuisanceParameterPull.evaluate_impact(model, minimizer,
                                                            nuip, nuip_hat - prefit_variation, rank_pois,
                                                            new_minimizer_options,  'tmp_snapshot')
                print('INFO: Time to find prefit down impact: {:.3f} s'.format(prefit_up_impact_time))
            else:
                print('WARNING: Prefit impact not estimated, instead postfit impact is cloned')
                pois_nom_up = [i for i in pois_up]
                pois_nom_down = [i for i in pois_down]
        else:
            pois_up, pois_down, pois_nom_up, pois_nom_down = [], [], [], []
        
        end_time = time.time()
        print('\nINFO: Time to perform all fits: {:.3f} s'.format(end_time-start_time))
        print('INFO: Unconditional minimum of NP {}: {} + {} - {}'.format(nuis_name, nuip_hat, 
              abs(nuip_errup), abs(nuip_errdown)))
        print('INFO: Prefit uncertainy of NP {}: {} +/- {}'.format(nuis_name, nuip_hat, prefit_variation))
        for i, rank_poi in enumerate(rank_pois):
            print('INFO: Unconditional minimum of POI {}: {}'.format(rank_poi.GetName(), pois_hat[i]))
            print('INFO: POI when varying NP up by 1 sigma postfit (prefit): {} ({})'.format(pois_up[i], pois_nom_up[i]))
            print('INFO: POI when varying NP down by 1 sigma postfit (prefit): {} ({})'.format(pois_down[i], pois_nom_down[i]))
            
        # store result in root file
        outname_root = os.path.join(outdir, nuis_name + '.root')
        
        result = {}
        result['nuis'] = {  'nuisance'   : nuis_name,
                            'nuis_nom'   : nuip_nom,
                            'nuis_hat'   : nuip_hat,
                            'nuis_hi'    : nuip_errup,
                            'nuis_lo'    : nuip_errdown,
                            'nuis_prefit': prefit_variation}
        result['pois'] = {}
        for i, rank_poi in enumerate(rank_pois):
            name = rank_poi.GetName()
            result['pois'][name] = { 'hat'     : pois_hat[i],
                                     'up'      : pois_up[i],
                                     'down'    : pois_down[i],
                                     'up_nom'  : pois_nom_up[i],
                                     'down_nom': pois_nom_down[i]}
            
        result_for_root = {}
        result_for_root.update(result['nuis'])
        for k,v in result['pois'].items():
            buffer = {'{}_{}'.format(k, kk): vv for kk,vv in v.items()}
            result_for_root.update(buffer)
        r_file = ROOT.TFile(outname_root, "RECREATE")
        r_tree = ROOT.TTree("result", "result")
        root_utils.fill_branch(r_tree, result_for_root)
        r_file.Write()
        r_file.Close()
        print('INFO: Saved output to {}'.format(outname_root))
        outname_json = os.path.join(outdir, nuis_name + '.json')
        json.dump(result, open(outname_json, 'w'), indent=2)
        
        if logfile_path is not None:
            common_utils.restore_stdout()

    @common_utils.timer
    def run_pulls(self, input_file:str='workspace.root', data:str='combData', poi:str='', 
                  snapshot:str='nominalNuis', outdir:str='output', profile:str='',
                  fix:str='', workspace:Optional[str]=None, model_config:Optional[str]=None,
                  minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', num_cpu:int=1, 
                  binned:bool=1, precision:float=0.001, eps:float=1.0, log_level:str='INFO', eigen:bool=False, 
                  strategy:int=0, print_level:int=1, fix_cache:bool=True, fix_multi:bool=True, offset:bool=True, 
                  optimize:int=2, parameter:str="", max_calls:int=-1, max_iters:int=-1, batch_mode:bool=False, 
                  int_bin_precision:float=-1.,parallel:int=0, cache:bool=True, exclude:str='', **kwargs):
        
        # configure default minimizer options
        ExtendedMinimizer._configure_default_minimizer_options(minimizer_type, minimizer_algo,
            strategy, debug_mode=(log_level=="DEBUG"))
        
        nuis_list = ExtendedModel.get_nuisance_parameter_names(input_file, workspace, model_config)
        include_patterns = parameter.split(',')
        if parameter:
            nuis_to_include = []
            for nuis_name in nuis_list:
                # filter out nuisance parameters
                if any(fnmatch.fnmatch(nuis_name, include_pattern) for include_pattern in include_patterns):
                    nuis_to_include.append(nuis_name)
        else:
            nuis_to_include = nuis_list
        exclude_patterns = exclude.split(',')
        nuis_to_exclude = []
        if exclude_patterns:
            for nuis_name in nuis_list:
                if any(fnmatch.fnmatch(nuis_name, exclude_pattern) for exclude_pattern in exclude_patterns):
                    nuis_to_exclude.append(nuis_name)
        nuis_names = list(set(nuis_to_include) - set(nuis_to_exclude))
        
        if log_level == 'INFO':
            logfile_paths = [os.path.join(outdir, '{}.log'.format(nuis_name)) for nuis_name in nuis_names]
        else:
            logfile_paths = [None]*len(nuis_names)
            

        arguments = (repeat(input_file), repeat(data), repeat(snapshot), nuis_names, 
                     repeat(poi), repeat(fix), repeat(profile), repeat(binned),
                     repeat(workspace), repeat(model_config), repeat(minimizer_type), 
                     repeat(minimizer_algo), repeat(precision), repeat(eps), repeat(strategy),
                     repeat(print_level), repeat(num_cpu), repeat(offset), repeat(optimize), 
                     repeat(eigen), repeat(fix_cache), repeat(fix_multi), repeat(max_calls), 
                     repeat(max_iters), repeat(batch_mode), repeat(int_bin_precision),
                     repeat(outdir), repeat(cache), logfile_paths)
        common_utils.execute_multi_tasks(self._run_pulls, *arguments, parallel=parallel)

        
    @staticmethod
    def parse_root_result(fname):
        with uproot.open(fname) as file:
            result = root_utils.uproot_to_dict(file)
        return result


from typing import Union

import numpy as np

def pretty_float(val:Union[str, float])->Union[int, float]:
    if float(val).is_integer():
        return int(float(val))
    return float(val)

def pretty_value(val:Union[int, float])->Union[int, float]:
    if isinstance(val, float) and val.is_integer():
        return int(val)
    return val

def is_integer(s:str):
    if not s:
        return False
    if len(s) == 1:
        return s.isdigit()
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()

def array_swap(arr1:np.ndarray, arr2:np.ndarray, indices):
    arr1[indices], arr2[indices] = arr2[indices], arr1[indices]

def df_array_swap(df, col1:str, col2:str, indices=None):
    if indices is None:
        df.loc[:, col1], df.loc[:, col2] = df[col2], df[col1]
    else:
        df.loc[indices, col1], df.loc[indices, col2] = df[indices][col2], df[indices][col1]
        
def reorder_arrays(*arrays, descending:bool=True):
    if descending:
        if not (arrays[0].dtype.type in [np.string_, np.str_]):
            indices = np.argsort(-arrays[0])
        else:
            indices = np.argsort(arrays[0])[::-1]
    else:
        indices = np.argsort(arrays[0])
    for arr in arrays:
        arr[:] = arr[indices]    

def reverse_arrays(*arrays):
    for arr in arrays:
        arr[:] = arr[::-1] 
        
def ceildiv(a, b):
    return -(-a // b)

def approx_n_digit(val:float, default=5):
    s = str(val)
    if not s.replace('.','',1).isdigit():
        return default
    elif '.' in s:
        return len(s.split('.')[1])
    else:
        return 0

def str_encode_value(val, n_digit=None, formatted=True):
    # account for the case where val is negative zero
    if round(float(val), 8) == 0:
        val = 0.
    if n_digit is not None:
        val_str = '{{:.{}f}}'.format(n_digit).format(val)
        #if val_str == '-{{:.{}f}}'.format(n_digit).format(0):
        #    val_str = '{{:.{}f}}'.format(n_digit).format(0)
    else:
        val_str = str(val)
    
    if formatted:
        val_str = val_str.replace('.', 'p').replace('-', 'n')
    return val_str

def str_decode_value(val_str):
    val = float(val_str.replace('p','.').replace('n','-'))
    return val


import sys
import traceback
from typing import Optional, Dict, List
from contextlib import contextmanager

@contextmanager
def standard_log(log_path:Optional[str]=None):
    import ROOT
    try:
        if log_path is not None:
            sys.stdout = open(log_path, 'w')
            ROOT.gSystem.RedirectOutput(log_path)
        yield sys.stdout
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
    finally:
        # recover stdout
        if log_path is not None:
            sys.stdout.close()
        sys.stdout = sys.__stdout__
        ROOT.gROOT.ProcessLine('gSystem->RedirectOutput(0);')